import argparse
import datetime
import sys
from argparse import RawTextHelpFormatter
from logging import Logger
from typing import Any, Sequence, Type, TypeAlias

import numpy as np
import pandas as pd

from .. import __title__, __version__
from ..utils._cli import console_exclusive_info, create_logger, log_textbox
from ..utils.read.product.file_info import ProductDataFrame, get_product_infos
from ._constants import PROGRAM_DESCRIPTION, PROGRAM_NAME, PROGRAM_SETUP_INSTRUCTIONS
from ._create_search_requests import create_search_request_list
from ._eo_product import EOProduct, _DownloadResult
from ._eo_search_request import EOSearchRequest
from ._organize_data import organize_data
from ._parse import (
    parse_path_to_config,
    parse_path_to_data,
    parse_search_inputs,
    parse_selected_index,
)
from ._remove_old_logs import remove_old_logs
from ._run_downloads import run_downloads
from ._run_search_requests import run_search_requets
from ._types import Entrypoint, _SearchInputs

RadiusMetersFloat: TypeAlias = float
LatFloat: TypeAlias = float
LonFloat: TypeAlias = float

LatSFloat: TypeAlias = float
LonWFloat: TypeAlias = float
LatNFloat: TypeAlias = float
LonEFloat: TypeAlias = float


def ecdownload(
    file_type: str | list[str],
    baseline: str | None = None,
    orbit_number: int | list[int] | None = None,
    start_orbit_number: int | None = None,
    end_orbit_number: int | None = None,
    frame_id: str | list[str] | None = None,
    orbit_and_frame: str | list[str] | None = None,
    start_orbit_and_frame: str | None = None,
    end_orbit_and_frame: str | None = None,
    timestamps: str | list[str] | None = None,
    start_time: str | None = None,
    end_time: str | None = None,
    radius_search: tuple[RadiusMetersFloat, LatFloat, LonFloat] | list | None = None,
    bounding_box: (
        tuple[LatSFloat, LonWFloat, LatNFloat, LonEFloat] | list | None
    ) = None,
    path_to_config: str | None = None,
    path_to_data: str | None = None,
    is_log: bool = False,
    is_debug: bool = False,
    is_download: bool = True,
    is_overwrite: bool = False,
    is_unzip: bool = True,
    is_delete: bool = True,
    is_create_subdirs: bool = True,
    is_export_results: bool = False,
    idx_selected_input: int | None = None,
    is_organize_data: bool = False,
    is_include_header: bool | None = None,
    is_reversed_order: bool = False,
    return_results: bool = False,
    verbose: bool = True,
) -> ProductDataFrame | None:
    """
    EarthCARE Download Tool: Search for and download EarthCARE products from a ESA data distribution platform (OADS or MAAP).

    The execution of this tool is divided into two parts:

    - First, based on provided arguments search request will be send via the OpenSearch API of the [ESA MAAP catalogue](https://catalog.maap.eo.esa.int/catalogue/).
    - Second, the resulting list of products is then downloaded from the configures download backend (OADS or MAAP). See:
        - MAAP: [portal.maap.eo.esa.int/earthcare](https://portal.maap.eo.esa.int/earthcare/)
        - OADS L1: [ec-pdgs-dissemination1.eo.esa.int](https://ec-pdgs-dissemination1.eo.esa.int/)
        - OADS L2: [ec-pdgs-dissemination2.eo.esa.int](https://ec-pdgs-dissemination2.eo.esa.int/)

    Args:
        file_type (str | list[str]): Name(s) of EarthCARE product(s) to search for (e.g., "ATL_NOM_1B", "ANOM", or "A-NOM").
            Note: Input string evaluation is not case sensitive. Also, product version may also be selected
            by adding a colon and the two-letter processor baseline after the name (e.g., "ANOM:BA").
        baseline (str | None, optional): Two-letter processor baseline used as default for all given `file_type`s (e.g., "BA").
            Note: A baseline specified in `file_type` with colon notation (e.g., "ANOM:BA") overwrites the default `baseline`.
            Defaults to None.
        orbit_number (int | list[int] | None, optional):
            Specific orbit number(s) to search for (e.g., 981 or [1000, 5000, ...]). Defaults to None.
        start_orbit_number (int | None, optional):
            The lower limit of orbit numbers to search for (e.g., 5000). Defaults to None.
        end_orbit_number (int | None, optional):
            The upper limit of orbit numbers to search for (e.g., 5003). Defaults to None.
        frame_id (str | list[str] | None, optional):
            Frame ID letter(s) to search for (i.e., letters A to H). Defaults to None.
        orbit_and_frame (str | list[str] | None, optional):
            Orbit and frame string(s) to search for (e.g., "01234F" or ["1000A", "5000C", ...]). Defaults to None.
        start_orbit_and_frame (str | None, optional):
            The lower limit of orbit and frames to search for (e.g., "05000D"). Defaults to None.
        end_orbit_and_frame (str | None, optional):
            The upper limit of orbit and frames to search for (e.g., "05003C"). Defaults to None.
        timestamps (str | list[str] | None, optional):
            Search for data containing specific timestamp(s) (e.g. "2024-07-31 13:45" or "20240731T134500Z"). Defaults to None.
        start_time (str | None, optional):
            The lower time limit for the search. Defaults to None.
        end_time (str | None, optional):
            The upper time limit for the search. Defaults to None.
        radius_search (tuple[RadiusMetersFloat, LatFloat, LonFloat] | list | None, optional):
            A tuple containing a radius (meters) and a lat/lon point to perform a geo radius search (e.g., 25000 51.35 12.43, i.e.,
            <radius[m]> <lat> <lon>). Latitudes must be provided as degrees north and longitudes as degrees east. Defaults to None.
        bounding_box (tuple[LatSFloat, LonWFloat, LatNFloat, LonEFloat]  |  list  |  None, optional):
            A tuple containing the extent for a bounding box geo search (e.g., [14.9, 37.7, 14.99, 37.78],
            i.e., <latS> <lonW> <latN> <lonE>). Latitudes must be provided as degrees north and longitudes as degrees east.
            Defaults to None.
        path_to_config (str | None, optional):
            If provided, uses given config file instead of the default config. Defaults to None.
        path_to_data (str | None, optional):
            If provided, downloads data to the given folder instead of the one defined in the config file. Defaults to None.
        is_log (bool, optional):
            If True, creates a log file in a `/log` folder inside the current working directory. Defaults to False.
        is_debug (bool, optional):
            If True, shows debug logs in the console. Defaults to False.
        is_download (bool, optional):
            If False, skips download part, but still performs search requests via the data dissemination platform API. Defaults to True.
        is_overwrite (bool, optional):
            If True, downloads and overwrites files that already exist in the data directory instead of skipping them. Defaults to False.
        is_unzip (bool, optional): If False, skips file extraction for downloaded archives. Defaults to True.
        is_delete (bool, optional):
            If True, deletes downloaded archives after extraction (i.e., does not delete non-extracted archives). Defaults to True.
        is_create_subdirs (bool, optional):
            If True, places downloaded files in a sub-directory structure according to the template defined in the config file.
            Defaults to True.
        is_export_results (bool, optional):
            If True, creates a text file in the current working directory listing all search results. Defaults to False.
        idx_selected_input (int | None, optional):
            A number matching an index in the list of found files. If provided, only this single file will be downloaded.
            Defaults to None.
        is_organize_data (bool, optional):
            If True, does not search or download any data. Defaults to False.
        is_include_header (bool | None, optional):
            If True, the full archive is downloaded containing both HDF5 data file (`.h5`) and header data file (`.HDR`).
            If False, only the data file will be downloaded, speeding up the download time.
            Defaults to None.

            !!! caution
                This option only applies to MAAP. OADS will always download the full archive.

        is_reversed_order (bool, optional):
            If True, downloads data products in reversed order (from the latest to the earliest). Defaults to False.
        return_results (bool, optional):
            If True, returns the search results as a `ProductDataFrame`. Defaults to False.
        verbose (bool, optional):
            If False, does not print logs to the console and does not create log file. Defaults to True.

    Returns:
        results (ProductDataFrame | None):
            If `return_results=False`, the function has no return (i.e., None).
            If `return_results=True`, the function returns the search results.
    """
    time_start_script: pd.Timestamp = pd.Timestamp(
        datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    )
    time_end_script: pd.Timestamp
    execution_time: pd.Timedelta

    def _to_list(input: Any, _type: Type) -> list | None:
        if isinstance(input, _type):
            return [input]
        elif isinstance(input, list):
            return input
        else:
            return None

    _file_type: list[str] | None = _to_list(file_type, str)
    assert isinstance(_file_type, list)
    file_type = _file_type

    orbit_number = _to_list(orbit_number, int)
    frame_id = _to_list(frame_id, str)
    orbit_and_frame = _to_list(orbit_and_frame, str)
    timestamps = _to_list(timestamps, str)

    if isinstance(radius_search, tuple):
        radius_search = list(radius_search)

    if isinstance(bounding_box, tuple):
        bounding_box = list(bounding_box)

    idx_selected: int | None = parse_selected_index(idx_selected_input)

    logger: Logger | None = None
    if verbose:
        logger = create_logger(
            logger_name=PROGRAM_NAME,
            log_to_file=is_log,
            debug=is_debug,
        )
    if is_log:
        remove_old_logs(100, pd.Timedelta(days=30))

    log_textbox(
        f"EarthCARE Download Tool\n{__title__} {__version__}",
        logger=logger,
        is_mayor=True,
    )

    if logger and not is_organize_data:
        logger.info(f"# Settings")
        logger.info(f"# - {is_download=}")
        logger.info(f"# - {is_overwrite=}")
        logger.info(f"# - {is_unzip=}")
        logger.info(f"# - {is_delete=}")
        logger.info(f"# - {is_create_subdirs=}")
        logger.info(f"# - {is_log=}")
        logger.info(f"# - {is_debug=}")
        logger.info(f"# - {is_export_results=}")
        logger.info(f"# - {idx_selected_input=}")

    config = parse_path_to_config(path_to_config, logger=logger)
    path_to_data = parse_path_to_data(path_to_data, logger=logger)
    if isinstance(path_to_data, str):
        config.path_to_data = path_to_data

    if logger and not is_organize_data:
        logger.info(f"# - config_filepath=<{config.filepath}>")
        logger.info(f"# - data_dirpath=<{config.path_to_data}>")

    if is_organize_data:
        if logger:
            logger.info(f"# Organizing local data ...")
        performed_moves = organize_data(
            config=config,
            logger=logger,
        )
        time_end_script = pd.Timestamp(
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        )
        execution_time = time_end_script - time_start_script
        execution_time_str = str(execution_time).split()[-1]
        if logger:
            console_exclusive_info()
        _moved = len([pm for pm in performed_moves if pm.get("status") == "success"])
        _failed = len([pm for pm in performed_moves if pm.get("status") == "error"])
        _msg = [
            f"EXECUTION SUMMARY",
            "---",
            f"Time taken          {execution_time_str}",
            f"Moved files         {_moved}",
            f"Failed moves        {_failed}",
        ]
        log_textbox("\n".join(_msg), logger=logger, show_time=True)
        return None

    if not isinstance(is_include_header, bool):
        is_include_header = config.maap_include_header_file

    search_inputs: _SearchInputs = parse_search_inputs(
        product_type=file_type,
        baseline=baseline,
        orbit_number=orbit_number,
        start_orbit_number=start_orbit_number,
        end_orbit_number=end_orbit_number,
        frame_id=frame_id,
        orbit_and_frame=orbit_and_frame,
        start_orbit_and_frame=start_orbit_and_frame,
        end_orbit_and_frame=end_orbit_and_frame,
        timestamps=timestamps,
        start_time=start_time,
        end_time=end_time,
        radius_search=radius_search,
        bounding_box=bounding_box,
        logger=logger,
    )
    if config.download_backend.lower() == "maap":
        entrypoint = Entrypoint.MAAP
    else:
        entrypoint = Entrypoint.OADS

    planned_requests: list[EOSearchRequest] = create_search_request_list(
        entrypoint=entrypoint,
        search_inputs=search_inputs,
        input_user_type=None,
        candidate_coll_names_user=[c.value for c in config.collections],
        logger=logger,
    )

    found_products: list[EOProduct] = run_search_requets(
        log_heading_msg=f"STEP 1/2 - Search products",
        search_requests=planned_requests,
        is_debug=is_debug,
        is_found_files_list_to_txt=is_export_results,
        selected_index=idx_selected,
        selected_index_input=idx_selected_input,
        logger=logger,
        download_only_h5=not is_include_header,
    )

    donwload_results: list[_DownloadResult] = run_downloads(
        log_heading_msg=f"STEP 2/2 - Download products",
        products=found_products,
        config=config,
        entrypoint=entrypoint,
        is_download=is_download,
        is_overwrite=is_overwrite,
        is_unzip=is_unzip,
        is_delete=is_delete,
        is_create_subdirs=is_create_subdirs,
        logger=logger,
        is_reversed_order=is_reversed_order,
    )

    if logger:
        num_downloads: int = 0
        num_unzips: int = 0
        num_errors: int = 0
        size_msg: str = "<missing size_msg>"
        avg_speed_mbs: float = 0.0
        if len(donwload_results) > 0:
            num_errors = sum([not r.success for r in donwload_results])
            num_downloads = sum([r.downloaded for r in donwload_results])
            num_unzips = sum([r.unzipped for r in donwload_results])
            total_size_mb = sum([r.size_mb for r in donwload_results])
            size_msg = f"{total_size_mb:.2f} MB"
            if total_size_mb >= 1024:
                size_msg = f"{total_size_mb / 1024:.2f} GB"
            avg_speed_mbs = float(np.mean([r.speed_mbs for r in donwload_results]))

        time_end_script = pd.Timestamp(
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        )
        execution_time = time_end_script - time_start_script
        execution_time_str = str(execution_time).split()[-1]

        console_exclusive_info()
        _msg = [
            f"EXECUTION SUMMARY",
            "---",
            f"Time taken          {execution_time_str}",
            f"API search requests {len(planned_requests)}",
            f"Remote files found  {len(found_products)}",
            f"Files downloaded    {num_downloads} ({size_msg} at ~{avg_speed_mbs:.2f} MB/s)",
            f"Files unzipped      {num_unzips}",
            f"Errors occured      {num_errors}",
        ]
        log_textbox("\n".join(_msg), logger=logger, show_time=True)

    if return_results:
        return get_product_infos([p.name for p in found_products], must_exist=False)
    return None


def cli_tool_ecdownload() -> None:
    parser = argparse.ArgumentParser(
        prog=PROGRAM_NAME,
        description=f"{PROGRAM_DESCRIPTION}\n\n{PROGRAM_SETUP_INSTRUCTIONS}",
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "product_type",
        type=str,
        nargs="*",
        help="A list of EarthCARE product names (e.g. ANOM or ATL-NOM-1B, etc.).\nYou can also specify the product version by adding a colon and the two-letter\nprocessor baseline after the name (e.g. ANOM:AD).",
    )
    parser.add_argument(
        "-d",
        "--path_to_data",
        "--path-to-data",
        type=str,
        default=None,
        help="The local root directory where products will be downloaded to",
    )
    parser.add_argument(
        "-o",
        "--orbit_number",
        "--orbit-number",
        type=int,
        nargs="*",
        default=None,
        help="A list of EarthCARE orbit numbers (e.g. 981)",
    )
    parser.add_argument(
        "-so",
        "--start_orbit_number",
        "--start-orbit-number",
        type=int,
        default=None,
        help="Start of orbit number range (e.g. 981). Can only be used in combination with option -eo.",
    )
    parser.add_argument(
        "-eo",
        "--end_orbit_number",
        "--end-orbit-number",
        type=int,
        default=None,
        help="End of orbit number range (e.g. 986). Can only be used in combination with option -so.",
    )
    parser.add_argument(
        "-f",
        "--frame_id",
        "--frame-id",
        type=str,
        nargs="*",
        default=None,
        help="A EarthCARE frame ID (i.e. single letters from A to H)",
    )
    parser.add_argument(
        "-oaf",
        "--orbit_and_frame",
        "--orbit-and-frame",
        type=str,
        nargs="*",
        default=None,
        help="A string describing the EarthCARE orbit number and frame (e.g. 00981E)",
    )
    parser.add_argument(
        "-soaf",
        "--start_orbit_and_frame",
        "--start-orbit-and-frame",
        type=str,
        default=None,
        help="Start orbit number and frame range (e.g. 00981E). Can only be used in combination with option -eoaf. Can not be used with separate orbit and frame options -o, -so, eo and -f.",
    )
    parser.add_argument(
        "-eoaf",
        "--end_orbit_and_frame",
        "--end-orbit-and-frame",
        type=str,
        default=None,
        help="End orbit number and frame range (e.g. 00982B). Can only be used in combination with option -soaf. Can not be used with separate orbit and frame options -o, -so, eo and -f.",
    )
    parser.add_argument(
        "-t",
        "--time",
        type=str,
        nargs="*",
        default=None,
        help='Search for data containing a specific timestamp (e.g. "2024-07-31 13:45" or 20240731T134500Z)',
    )
    parser.add_argument(
        "-st",
        "--start_time",
        "--start-time",
        type=str,
        default=None,
        help='Start of sensing time (e.g. "2024-07-31 13:45" or 20240731T134500Z)',
    )
    parser.add_argument(
        "-et",
        "--end_time",
        "--end-time",
        type=str,
        default=None,
        help='End of sensing time (e.g. "2024-07-31 13:45" or 20240731T134500Z)',
    )
    parser.add_argument(
        "-r",
        "--radius_search",
        "--radius-search",
        type=float,
        nargs=3,
        default=None,
        help="Perform search around a radius around a point (e.g. 25000 51.35 12.43, i.e. <radius[m]> <latitude> <longitude>)",
    )
    parser.add_argument(
        "-pv",
        "--product_version",
        "--product-version",
        "--baseline",
        type=str,
        default="latest",
        help='Product version, i.e. the two-letter identifier of the processor baseline (e.g. AC). Defalut ist "latest"',
    )
    parser.add_argument(
        "-bbox",
        "--bounding_box",
        "--bounding-box",
        type=float,
        nargs=4,
        default=None,
        help="Perform search inside a bounding box (e.g. 14.9 37.7 14.99 37.78, i.e. <latS> <lonW> <latN> <lonE>)",
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite local data (otherwise existing local data will not be downloaded again)",
    )
    parser.add_argument(
        "--no_download",
        "--no-download",
        action="store_false",
        help="Do not download any data",
    )
    parser.add_argument(
        "--no_unzip",
        "--no-unzip",
        action="store_false",
        help="Do not unzip any data",
    )
    parser.add_argument(
        "--no_delete",
        "--no-delete",
        action="store_false",
        help="Do not delete zip files after unzipping them",
    )
    parser.add_argument(
        "--no_subdirs",
        "--no-subdirs",
        action="store_false",
        help="Do not create subdirs like: data_directory/data_level/product_type/year/month/day",
    )
    parser.add_argument(
        "-c",
        "--path_to_config",
        "--path-to-config",
        type=str,
        default=None,
        help="The path to an OADS credential TOML file (note: if not provided, a file named 'config.toml' is required in the script's folder)",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Shows debug messages in console.",
    )
    parser.add_argument(
        "--no_log",
        "--no-log",
        action="store_false",
        help="Prevents generation of log files.",
    )
    parser.add_argument(
        "-i",
        "--select_file_at_index",
        "--select-file-at-index",
        type=int,
        default=None,
        help="Select only one product from the found products list by index for download. You may provide a negative index to start from the last entry (e.g. -1 downloads the last file listed).",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="store_true",
        help="Shows the script's version and exit",
    )
    parser.add_argument(
        "--export_results",
        "--export-results",
        action="store_true",
        help="Writes names of found files to a txt file called 'results.txt'",
    )
    parser.add_argument(
        "--organize_data",
        "--organize-data",
        action="store_true",
        help="Ensures that all EarthCARE data products under your data folder are located correctly in the subfolder structure. When this option is used, no data will be downloaded, only local data folders will be moved if necessary.",
    )
    parser.add_argument(
        "--include_header",
        "--include-header",
        action="store_true",
        help="Includes header file (.HDR) in product download when using MAAP",
    )
    parser.add_argument(
        "--exclude_header",
        "--exclude-header",
        action="store_true",
        help="Does not download header file (.HDR) but only the product's .h5-file when using MAAP",
    )
    parser.add_argument(
        "--reversed_order",
        "--reversed-order",
        action="store_true",
        help="Downloads data products in reversed order (from the latest to the earliest)",
    )
    args = parser.parse_args()

    if args.version:
        print(f"earthcarekit {__version__}")
        sys.exit(0)

    path_to_config = args.path_to_config
    path_to_data = args.path_to_data

    is_overwrite: bool = args.overwrite
    is_download: bool = args.no_download
    is_unzip: bool = args.no_unzip
    is_delete: bool = args.no_delete
    is_create_subdirs: bool = args.no_subdirs
    is_log: bool = args.no_log
    is_debug: bool = args.debug
    idx_selected_input: int | None = args.select_file_at_index
    idx_selected: int | None = parse_selected_index(args.select_file_at_index)
    is_export_results: bool = args.export_results
    is_organize_data: bool = args.organize_data

    product_type: list[str] = args.product_type
    product_version: str = args.product_version
    orbit_number: list[int] | None = args.orbit_number
    start_orbit_number: int | None = args.start_orbit_number
    end_orbit_number: int | None = args.end_orbit_number
    frame_id: list[str] | None = args.frame_id
    orbit_and_frame: list[str] | None = args.orbit_and_frame
    start_orbit_and_frame: str | None = args.start_orbit_and_frame
    end_orbit_and_frame: str | None = args.end_orbit_and_frame
    timestamps: list[str] | None = args.time
    start_time: str | None = args.start_time
    end_time: str | None = args.end_time
    radius_search: list[float] | None = args.radius_search
    bounding_box: list[float] | None = args.bounding_box
    include_header: bool = args.include_header
    exclude_header: bool = args.exclude_header
    reversed_order: bool = args.reversed_order

    is_include_header: bool | None = None
    if include_header and exclude_header:
        print(
            f"You can't use options '--include_header' and '--exclude_header' together."
        )
        sys.exit(0)
    elif include_header:
        is_include_header = True
    elif exclude_header:
        is_include_header = False

    ecdownload(
        file_type=product_type,
        baseline=product_version,
        orbit_number=orbit_number,
        start_orbit_number=start_orbit_number,
        end_orbit_number=end_orbit_number,
        frame_id=frame_id,
        orbit_and_frame=orbit_and_frame,
        start_orbit_and_frame=start_orbit_and_frame,
        end_orbit_and_frame=end_orbit_and_frame,
        timestamps=timestamps,
        start_time=start_time,
        end_time=end_time,
        radius_search=radius_search,
        bounding_box=bounding_box,
        path_to_config=path_to_config,
        path_to_data=path_to_data,
        is_log=is_log,
        is_debug=is_debug,
        is_download=is_download,
        is_overwrite=is_overwrite,
        is_unzip=is_unzip,
        is_delete=is_delete,
        is_create_subdirs=is_create_subdirs,
        is_export_results=is_export_results,
        idx_selected_input=idx_selected_input,
        is_organize_data=is_organize_data,
        is_include_header=is_include_header,
        is_reversed_order=reversed_order,
    )


def main() -> None:
    cli_tool_ecdownload()


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from logging import Logger
from typing import Literal, Sequence

import numpy as np
import xarray as xr
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ...utils.time import TimestampLike
from ..figure import ECKFigure
from ..save import save_plot


@dataclass
class QuicklookFigure:
    """Container for quicklook figure results.

    Attributes:
        fig (Figure): The main figure object (`matplotlib.figure.Figure`) that contains all subfigures (`matplotlib.figure.SubFigure`).
        subfigs (list[list[ECKFigure]]):
            Nested list of earthcarekik figure objects, where each inner list groups
            subfigures belonging to a particular panel, section, or layout element.
    """

    fig: Figure
    subfigs: list[list[ECKFigure]]

    def save(
        self,
        filename: str = "",
        filepath: str | None = None,
        ds: xr.Dataset | None = None,
        ds_filepath: str | None = None,
        dpi: float | Literal["figure"] = "figure",
        orbit_and_frame: str | None = None,
        utc_timestamp: TimestampLike | None = None,
        use_utc_creation_timestamp: bool = False,
        site_name: str | None = None,
        hmax: int | float | None = None,
        radius: int | float | None = None,
        extra: str | None = None,
        transparent_outside: bool = False,
        verbose: bool = True,
        print_prefix: str = "",
        create_dirs: bool = False,
        transparent_background: bool = False,
        resolution: str | None = None,
        **kwargs,
    ) -> None:
        """
        Save a figure as an image or vector graphic to a file and optionally format the file name in a structured way using EarthCARE metadata.

        Args:
            figure (Figure | HasFigure): A figure object (`matplotlib.figure.Figure`) or objects exposing a `.fig` attribute containing a figure (e.g., `CurtainFigure`).
            filename (str, optional): The base name of the file. Can be extended based on other metadata provided. Defaults to empty string.
            filepath (str | None, optional): The path where the image is saved. Can be extended based on other metadata provided. Defaults to None.
            ds (xr.Dataset | None, optional): A EarthCARE dataset from which metadata will be taken. Defaults to None.
            ds_filepath (str | None, optional): A path to a EarthCARE product from which metadata will be taken. Defaults to None.
            pad (float, optional): Extra padding (i.e., empty space) around the image in inches. Defaults to 0.1.
            dpi (float | 'figure', optional): The resolution in dots per inch. If 'figure', use the figure's dpi value. Defaults to None.
            orbit_and_frame (str | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            utc_timestamp (TimestampLike | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            use_utc_creation_timestamp (bool, optional): Whether the time of image creation should be included in the file name. Defaults to False.
            site_name (str | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            hmax (int | float | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            radius (int | float | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            resolution (str | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            extra (str | None, optional): A custom string to be included in the file name. Defaults to None.
            transparent_outside (bool, optional): Whether the area outside figures should be transparent. Defaults to False.
            verbose (bool, optional): Whether the progress of image creation should be printed to the console. Defaults to True.
            print_prefix (str, optional): A prefix string to all console messages. Defaults to "".
            create_dirs (bool, optional): Whether images should be saved in a folder structure based on provided metadata. Defaults to False.
            transparent_background (bool, optional): Whether the background inside and outside of figures should be transparent. Defaults to False.
            **kwargs (dict[str, Any]): Keyword arguments passed to wrapped function call of `matplotlib.pyplot.savefig`.
        """
        save_plot(
            fig=self.fig,
            filename=filename,
            filepath=filepath,
            ds=ds,
            ds_filepath=ds_filepath,
            dpi=dpi,
            orbit_and_frame=orbit_and_frame,
            utc_timestamp=utc_timestamp,
            use_utc_creation_timestamp=use_utc_creation_timestamp,
            site_name=site_name,
            hmax=hmax,
            radius=radius,
            extra=extra,
            transparent_outside=transparent_outside,
            verbose=verbose,
            print_prefix=print_prefix,
            create_dirs=create_dirs,
            transparent_background=transparent_background,
            resolution=resolution,
            **kwargs,
        )



from contextlib import nullcontext
from typing import Sequence

import numpy as np
from numpy.typing import NDArray
from xarray import Dataset

from ...utils.geo import geodesic, get_coords
from ...utils.read import read_product, read_products, trim_to_latitude_frame_bounds
from ...utils.read.product._rebin_xmet_to_vertical_track import (
    rebin_xmet_to_vertical_track,
)
from ...utils.time import TimeRangeLike
from ...utils.typing import DistanceRangeLike
from ...utils.xarray_utils import concat_datasets, filter_time
from ..figure import CurtainFigure, ECKFigure, FigureType, MapFigure, SwathFigure
from ..figure.multi_panel import create_multi_figure_layout
from ._quicklook_results import QuicklookFigure


def ecquicklook_psc(
    anom: str | Sequence[str] | NDArray[np.str_] | Dataset,
    xmet: str | Sequence[str] | NDArray[np.str_] | Dataset | None = None,
    zoom_at: float | None = 0.5,
    height_range: DistanceRangeLike | None = (0, 40e3),
    time_range: TimeRangeLike | None = None,
    info_text_loc: str | None = None,
) -> QuicklookFigure:
    """
    Creates a two-column multi-panel quicklook of a PSC event, displaying:

    - 1st column: Two maps showing the EarthCARE track.
    - 2nd column: Three rows showing co- and cross-polar attenuated backscatter and the calculated depolarization ratio.

    Args:
        anom (str | Sequence[str] | Dataset):  The ATL_NOM_1B product filepath(s) or dataset(s).
        xmet (str | Sequence[str] | Dataset | None, optional): The AUX_MET_1D product filepath(s) or dataset(s).
            If given, temperature contour lines will be added to the plots. Defaults to None.
        zoom_at (float | None, optional): In case two frames are given, selects only a zoomed-in portion of the
            frames around this fractional index (0 -> only 1st frame, 0.5 -> half of end of 1st and half of beginning
            of 2nd frame, 1 -> only 2nd frame). Defaults to 0.5.
        height_range (DistanceRangeLike | None, optional): _description_. Defaults to (0, 40e3).
        time_range (TimeRangeLike | None, optional): A time range to filter the displayed data. Defaults to None.
        info_text_loc (str | None, optional): The positioning of the orbt, frame and product info text (e.g., "upper right").
            Defaults to None.

    Raises:
        ValueError: If none or more than 2 frames are given.
        ValueError: If given number X-MET files does not match number of A-NOM files.

    Returns:
        QuicklookFigure: The quicklook object.

    Examples:
        ```python
        import earthcarekit as eck

        df = eck.search_product(
            file_type=["anom", "xmet"],
            orbit_and_frame=["3579B", "3579C"],
        ).filter_latest()

        fps_anom = df.filter_file_type("anom").filepath
        fps_xmet = df.filter_file_type("xmet").filepath

        ql = eck.ecquicklook_psc(
            anom=fps_anom,
            xmet=fps_xmet,
        )
        ```

        ![ecquicklook_psc.png](https://raw.githubusercontent.com/TROPOS-RSD/earthcarekit-docs-assets/refs/heads/main/assets/images/quicklooks/ecquicklook_psc.png)
    """

    if not isinstance(anom, str) and isinstance(anom, (Sequence, np.ndarray)):
        if len(anom) == 0 or len(anom) > 2:
            raise ValueError(
                f"supports input of either 1 or 2 consecutive frames, but got {len(anom)} A-NOM frames"
            )
        if not isinstance(xmet, str) and isinstance(xmet, (Sequence, np.ndarray)):
            if len(anom) != len(xmet):
                raise ValueError(
                    f"number of X-MET frames ({len(xmet)}) must match number of A-NOM frames ({len(anom)})"
                )

    def _load_full_anom() -> Dataset:
        if isinstance(anom, Dataset):
            return anom
        elif isinstance(anom, str):
            return read_product(anom)
        return read_products(anom)

    def _load_anom() -> Dataset:
        if isinstance(anom, Dataset):
            return anom
        elif isinstance(anom, str):
            return read_product(anom)
        return read_products(anom, zoom_at=zoom_at)

    def _load_xmet() -> Dataset | None:
        if isinstance(xmet, Dataset):
            return xmet
        elif isinstance(xmet, str):
            return read_product(xmet)
        elif isinstance(xmet, (Sequence, np.ndarray)) and len(xmet) > 0:
            return read_product(xmet[0])
        return None

    def _load_xmet2() -> Dataset | None:
        if (
            not isinstance(xmet, Dataset)
            and not isinstance(xmet, str)
            and isinstance(xmet, (Sequence, np.ndarray))
            and len(xmet) > 1
        ):
            return read_product(xmet[1])
        return None

    with (
        _load_full_anom() as ds_full,
        _load_anom() as ds,
        nullcontext(_load_xmet()) as ds_xmet,
        nullcontext(_load_xmet2()) as ds_xmet2,
    ):
        if isinstance(ds_xmet, Dataset):
            ds_xmet = rebin_xmet_to_vertical_track(ds_xmet, ds_full)
            ds_xmet = trim_to_latitude_frame_bounds(ds_xmet)

            if isinstance(ds_xmet2, Dataset):
                ds_xmet2 = rebin_xmet_to_vertical_track(ds_xmet2, ds_full)
                ds_xmet2 = trim_to_latitude_frame_bounds(ds_xmet2)

                ds_xmet = concat_datasets(ds_xmet, ds_xmet2, "along_track")

        ds = filter_time(ds, time_range)

        figs: list[list[ECKFigure]] = [[], []]
        layout = create_multi_figure_layout(
            rows=[
                FigureType.CURTAIN,
                FigureType.CURTAIN,
                FigureType.CURTAIN,
            ],
            hspace=0.4,
            wspace=1.0,
            map_rows=[
                FigureType.MAP_1_ROW,
                FigureType.MAP_2_ROW,
            ],
        )

        mf1 = MapFigure(
            ax=layout.axs_map[0],
            show_grid_labels=False,
        )
        mf1.ecplot(ds)
        mf1.plot_track(
            latitude=ds_full.latitude.values,
            longitude=ds_full.longitude.values,
            color="white",
            linestyle="dashed",
            linewidth=1,
            zorder=2,
            highlight_first=False,
            highlight_last=False,
            alpha=0.8,
        )
        figs[0].append(mf1)

        mf2 = MapFigure(
            ax=layout.axs_map[1],
            show_top_labels=False,
            show_right_labels=False,
            style="blue_marble",
            coastlines_resolution="50m",
            show_text_time=False,
            show_text_frame=False,
        )
        mf2.ecplot(ds)
        mf2.plot_track(
            latitude=ds_full.latitude.values,
            longitude=ds_full.longitude.values,
            color="white",
            linestyle="dashed",
            linewidth=1,
            zorder=2,
            highlight_first=False,
            highlight_last=False,
            alpha=0.8,
        )

        coords = get_coords(ds)
        zoom_radius_meters = geodesic(coords[0], coords[-1], units="m") * 0.4
        mf2.ax.set_xlim(-zoom_radius_meters, zoom_radius_meters)  # type: ignore
        mf2.ax.set_ylim(-zoom_radius_meters, zoom_radius_meters)  # type: ignore
        figs[0].append(mf2)

        vars = [
            "mie_attenuated_backscatter",
            "crosspolar_attenuated_backscatter",
            "depol_ratio",
        ]

        for i, (var, ax) in enumerate(zip(vars, layout.axs)):
            if i == 0:
                ax_style_top = "utc"
                ax_style_bottom = "lat"
            elif i == 1:
                ax_style_top = "lat_nolabels"
                ax_style_bottom = "lon"
            elif i == 2:
                ax_style_top = "lon_nolabels"
                ax_style_bottom = "distance"

            cf = CurtainFigure(
                ax=ax,
                ax_style_top=ax_style_top,
                ax_style_bottom=ax_style_bottom,
                num_ticks=8,
            )
            cf.ecplot(
                ds=ds,
                var=var,
                label_length=55,
                height_range=height_range,
                info_text_loc=info_text_loc,
            )
            cf.ecplot_temperature(
                ds=ds,
                colors="#fffffff0",
                levels=[-90, -80, -60, -40, -20, 0, 10],
                label_levels=[-90, -80, -60, -40, -20, 0, 10],
                linestyles="solid",
                linewidths=[1, 0.7, 0.5, 1, 0.5, 1, 0.5],
            )
            if isinstance(ds_xmet, Dataset):
                cf.ecplot_tropopause(ds_xmet)
            figs[1].append(cf)

        return QuicklookFigure(
            fig=layout.fig,
            subfigs=figs,
        )



from typing import Literal

import numpy as np
import xarray as xr
from numpy.typing import NDArray

from ..constants import ALONG_TRACK_DIM, TRACK_LAT_VAR, TRACK_LON_VAR
from ..geo import haversine
from ..geo import vincenty as geodesic
from ..geo.coordinates import get_coords
from ..ground_sites import GroundSite, get_ground_site
from ..np_array_utils import pad_true_sequence
from .exception import EmptyFilterResultError


def filter_radius(
    ds: xr.Dataset,
    # *,
    radius_km: float,
    center_lat: float | None = None,
    center_lon: float | None = None,
    site: GroundSite | str | None = None,
    lat_var: str = TRACK_LAT_VAR,
    lon_var: str = TRACK_LON_VAR,
    along_track_dim: str = ALONG_TRACK_DIM,
    method: Literal["geodesic", "haversine"] = "geodesic",
    closest: bool = False,
    pad_idxs: int = 0,
) -> xr.Dataset:
    """
    Filters a dataset to include only points within a specified radius of a geographic location.

    Args:
        ds (xr.Dataset): Input dataset with geolocation data.
        radius_km (float): Radius (in kilometers) around the center location.
        site (GroundSite or str, optional): GroundSite object or name from which center location will be retrieved,
            alternatively `center_lat` and `center_lon` must be set.
        center_lat (float, optional): Latitude of the center point,
            alternatively `site` must be set.
        center_lon (float, optional): Longitude of the center point,
            alternatively `site` must be set.
        lat_var (str, optional): Name of the latitude variable. Defaults to TRACK_LAT_VAR.
        lon_var (str, optional): Name of the longitude variable. Defaults to TRACK_LON_VAR.
        along_track_dim (str, optional): Dimension along which to apply filtering. Defaults to ALONG_TRACK_DIM.
        method (Literal["geodesic", "haversine"], optional): Distance calculation method. Defaults to "geodesic".
        closest (bool, optional): If True, only the single closest sample is returned, otherwise all samples within radius. Defaults to False.
        pad_idxs (int, optional): Number of additional samples added at both sides of the selection. Defaults to 0.

    Returns:
        xr.Dataset: Filtered dataset containing only points within the specified radius.

    Raises:
        EmptyFilterResultError: If no data points are found within the radius.
        ValueError: If the `method` is invalid.
    """
    _center_lat: float
    _center_lon: float

    if isinstance(site, str):
        site = get_ground_site(site)

    if isinstance(site, GroundSite):
        _center_lat = site.latitude
        _center_lon = site.longitude
    elif isinstance(center_lat, (int, float, np.integer, np.floating)) and isinstance(
        center_lon, (int, float, np.integer, np.floating)
    ):
        _center_lat = float(center_lat)
        _center_lon = float(center_lon)
    else:
        raise ValueError(
            f"Either 'site' or 'center_lat' and 'center_lon' must be given."
        )

    if method not in ["geodesic", "haversine"]:
        raise ValueError(
            r'Invalid method choosen. Available methods: {"geodesic", "haversine"}'
        )

    satellite_coords = get_coords(ds, lat_var=lat_var, lon_var=lon_var)

    center_coords = (_center_lat, _center_lon)

    if method == "geodesic":
        distances = geodesic(center_coords, satellite_coords)
    else:
        distances = haversine(center_coords, satellite_coords)

    mask = np.array(distances < radius_km)

    if closest:
        closest_distance = np.min(distances)
        closest_filtered_index = int(np.argmin(np.abs(distances - closest_distance)))
        mask[:] = False
        mask[closest_filtered_index] = True

    mask = pad_true_sequence(mask, pad_idxs)

    da_mask = xr.DataArray(data=mask, dims=[along_track_dim])
    if np.sum(da_mask.values) < 1:
        raise EmptyFilterResultError(
            f"Could not find valid overpass for given inputs. Data lies outside the given {radius_km} km radius around ({center_lat} degN {center_lon} degE).",
            min_distance=float(np.min(distances)),
        )

    ds_new: xr.Dataset = xr.Dataset(
        {
            var: (
                ds[var].copy().where(da_mask, drop=True)
                if along_track_dim in ds[var].dims
                else ds[var].copy()
            )
            for var in ds.data_vars
        }
    )
    ds_new.attrs = ds.attrs.copy()
    ds_new.encoding = ds.encoding.copy()

    return ds_new



import numpy as np
import xarray as xr
from xarray import Dataset


def pad_dataset(ds: Dataset, target_sizes: dict, int_fill: int = -9999) -> Dataset:
    """Pad a dataset to match target sizes along all relevant dimensions."""
    padded_vars = {}
    for name, var in ds.data_vars.items():
        pad_width = []
        for dim in var.dims:
            current = ds.sizes.get(dim, 0)
            target = target_sizes.get(dim, current)
            pad = (0, target - current) if target > current else (0, 0)
            pad_width.append(pad)

        if not pad_width:
            continue  # Skip scalar variables

        fill_value: int | float
        if np.issubdtype(var.dtype, np.integer):
            if np.iinfo(var.dtype).min <= int_fill:
                fill_value = int_fill
            else:
                fill_value = np.iinfo(var.dtype).min
        else:
            fill_value = np.nan

        padded_values = np.pad(var.values, pad_width, constant_values=fill_value)
        padded_vars[name] = xr.DataArray(padded_values, dims=var.dims, attrs=var.attrs)

    return Dataset(padded_vars, attrs=ds.attrs)


def concat_datasets(ds1: Dataset, ds2: Dataset, dim: str) -> Dataset:
    """Concatenate two `xarray.Dataset` objects along a specified dimension, padding other dimensions to match.

    Pads all non-concatenation dimensions in both datasets to the maximum size among them
    (if they differ) before concatenating. Integer variables are padded with -9999 or data
    type-specific minimum value (e.g., -128 for int8), non-interger variables are padded with NaN.

    Args:
        ds1 (Dataset): The first dataset to concatenate.
        ds2 (Dataset): The second dataset to concatenate.
        dim (str): The name of the dimension to concatenate along.

    Returns:
        Dataset: A new dataset resulting from the concatenation.
    """

    def get_scalars(ds: xr.Dataset) -> list:
        scalars = [k for k, v in ds.data_vars.items() if v.ndim == 0]
        return scalars

    ds1_scalars = get_scalars(ds1)
    ds2_scalars = get_scalars(ds2)
    scalar_vars: list = list(set(ds1_scalars + ds2_scalars))

    scalar_data: dict = {v: [] for v in scalar_vars}
    for v in scalar_vars:
        if v in ds1:
            scalar_data[v].extend(np.atleast_1d(ds1[v].values))
        if v in ds2:
            scalar_data[v].extend(np.atleast_1d(ds2[v].values))

    max_dim_sizes = {
        d: max(ds1.sizes.get(d, 0), ds2.sizes.get(d, 0))
        for d in set(ds1.dims).union(ds2.dims)
        if d != dim
    }

    ds1_padded = pad_dataset(ds1, max_dim_sizes)
    ds2_padded = pad_dataset(ds2, max_dim_sizes)

    ds_combined = xr.concat([ds1_padded, ds2_padded], dim=dim)

    if "concat_dim" in ds_combined.dims:
        ds_combined = ds_combined.drop_dims("concat_dim", errors="ignore")

    for v in scalar_vars:
        da = xr.DataArray(scalar_data[v], dims=["concat_dim"])
        ds_combined[v] = da

    source1 = ds1.encoding.get("source")
    source2 = ds2.encoding.get("source")
    sources = [s for s in [source1, source2] if isinstance(s, str)]

    if len(sources) > 0:
        ds_combined.encoding["sources"] = sources

    return ds_combined



from ...utils.typing import DistanceRangeLike


def set_none_height_range_to_default(
    height_range: DistanceRangeLike | None, hmin: float, hmax: float
) -> tuple[float, float]:
    if height_range is None:
        return (hmin, hmax)

    new_height_range: tuple
    new_height_range = (height_range[0], height_range[1])
    if new_height_range[0] is None:
        new_height_range = (hmin, new_height_range[1])

    if new_height_range[1] is None:
        new_height_range = (new_height_range[0], hmax)

    return (float(new_height_range[0]), float(new_height_range[1]))



import numpy as np
from matplotlib.colors import Colormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex
from .cmap import Cmap


def get_cmap_cpr_hydrometeor_classification():
    cmap_data = [
        ("#929591", -1, "No data"),
        ("#7f2b0a", 0, "Surface"),
        ("#ffffff", 1, "Clear"),
        ("#ffff84", 2, "Liquid cloud"),
        ("#f5bf03", 3, "Drizzling liquid clouds"),
        ("#f97306", 4, "Warm rain"),
        ("#ff000d", 5, "Cold rain"),
        ("#c071fe", 6, "Melting snow"),
        ("#004577", 7, "Rimed snow"),
        ("#0165fc", 8, "Snow"),
        ("#95d0fc", 9, "Ice"),
        ("#d7fffe", 10, "Stratospheric ice"),
        ("#7a9703", 11, "Insects"),
        ("#840000", 12, "Heavy rain likely"),
        ("#0504aa", 13, "Mixed-phase precip. likely"),
        ("#840000", 14, "Heavy rain"),
        ("#001146", 15, "Heavy mixed-phase precip."),
        ("#bb3f3f", 16, "Rain in clutter"),
        ("#5684ae", 17, "Snow in clutter"),
        ("#eedc5b", 18, "Cloud in clutter"),
        ("#d8dcd6", 19, "Clear likely"),
        ("#c5c9c7", 20, "Uncertain"),
    ]

    colors = [c[0] for c in cmap_data]
    definitions = {c[1]: c[2] for c in cmap_data}

    cmap = Cmap(
        colors=colors,
        name="cpr_hydrometeor_classification",
    ).to_categorical(definitions)
    return cmap


def get_cmap_cpr_doppler_velocity_classification():
    cmap_data = [
        ("#929591", -1, "No data"),
        ("#7f2b0a", 0, "Surface"),
        ("#ffffff", 1, "Clear"),
        ("#ff84f9", 2, "Dominated by $V_t$"),
        ("#75acff", 3, "Dominated by $\mathit{w}$"),
        ("#8811be", 4, "Contribution by $V_t$ and $\mathit{w}$"),
        ("#c5c9c7", 5, "Uncertain"),
        ("#840000", 12, "Heavy rain likely"),
        ("#0504aa", 13, "Mixed-phase precip. likely"),
        ("#840000", 14, "Heavy rain"),
        ("#001146", 15, "Heavy mixed-phase precip."),
        ("#bb3f3f", 16, "Rain in clutter"),
        ("#5684ae", 17, "Snow in clutter"),
        ("#eedc5b", 18, "Cloud in clutter"),
        ("#d8dcd6", 19, "Clear likely"),
    ]

    colors = [c[0] for c in cmap_data]
    definitions = {c[1]: c[2] for c in cmap_data}

    cmap = Cmap(
        colors=colors,
        name="cpr_doppler_velocity_classification",
    ).to_categorical(definitions)
    return cmap


def get_cmap_cpr_simplified_convective_classification():
    cmap_data = [
        ("#929591", -1, "No data"),
        ("#7f2b0a", 0, "Surface"),
        ("#ffffff", 1, "Clear"),
        ("#ffb584", 2, "Weak conv. + stratiform clouds"),
        ("#66d2da", 3, "Deep conv. clouds"),
        ("#df54bc", 4, "Dynamic conv. cores"),
        ("#c5c9c7", 5, "Uncertain"),
        ("#840000", 12, "Heavy rain likely"),
        ("#0504aa", 13, "Mixed-phase precip. likely"),
        ("#840000", 14, "Heavy rain"),
        ("#001146", 15, "Heavy mixed-phase precip."),
        ("#bb3f3f", 16, "Rain in clutter"),
        ("#5684ae", 17, "Snow in clutter"),
        ("#eedc5b", 18, "Cloud in clutter"),
        ("#d8dcd6", 19, "Clear likely"),
    ]

    colors = [c[0] for c in cmap_data]
    definitions = {c[1]: c[2] for c in cmap_data}

    cmap = Cmap(
        colors=colors,
        name="cpr_simplified_convective_classification",
    ).to_categorical(definitions)
    return cmap



"""
**earthcarekit.read**

Reading utilities for EarthCARE product data.

---
"""

from ._read_any import read_any
from ._read_nc import read_nc
from ._read_polly import read_polly
from .product import (
    add_isccp_cloud_type,
    read_hdr_fixed_header,
    read_header_data,
    read_product,
    read_products,
    read_science_data,
    rebin_msi_to_jsg,
    search_product,
    trim_to_latitude_frame_bounds,
)
from .product._rebin_xmet_to_vertical_track import rebin_xmet_to_vertical_track
from .product.file_info import (
    FileAgency,
    FileLatency,
    FileMissionID,
    FileType,
    ProductInfo,
    get_file_type,
    get_product_info,
    get_product_infos,
    is_earthcare_product,
)
from .product.level1.atl_nom_1b import add_depol_ratio, add_scattering_ratio
from .search import search_files_by_regex

__all__ = [
    "read_hdr_fixed_header",
    "read_header_data",
    "read_product",
    "read_products",
    "read_science_data",
    "read_nc",
    "rebin_xmet_to_vertical_track",
    "rebin_msi_to_jsg",
    "trim_to_latitude_frame_bounds",
    "search_product",
    "FileAgency",
    "FileLatency",
    "FileMissionID",
    "FileType",
    "ProductInfo",
    "get_file_type",
    "get_product_info",
    "get_product_infos",
    "is_earthcare_product",
    "search_files_by_regex",
    "read_polly",
    "read_any",
    "add_depol_ratio",
    "add_scattering_ratio",
    "add_isccp_cloud_type",
]



from .atl_aer_2a import read_product_aaer
from .atl_ald_2a import read_product_aald
from .atl_cla_2a import read_product_acla
from .atl_cth_2a import read_product_acth
from .atl_ebd_2a import read_product_aebd
from .atl_fm__2a import read_product_afm
from .atl_ice_2a import read_product_aice
from .atl_tc__2a import read_product_atc
from .cpr_cd__2a import read_product_ccd
from .cpr_cld_2a import read_product_ccld
from .cpr_clp_2a import read_product_cclp
from .cpr_eco_2a import read_product_ceco
from .cpr_fmr_2a import read_product_cfmr
from .cpr_tc__2a import read_product_ctc
from .msi_cm__2a import read_product_mcm
from .msi_cop_2a import read_product_mcop



from ._get_day_night_mask import get_day_night_mask
from ._time import (
    TimedeltaLike,
    TimeRangeLike,
    TimestampComparisonResult,
    TimestampLike,
    check_if_same_timestamp,
    format_time_range_text,
    get_time_range,
    lookup_value_by_timestamp,
    num_to_time,
    time_to_iso,
    time_to_num,
    time_to_string,
    to_timedelta,
    to_timedeltas,
    to_timestamp,
    to_timestamps,
    validate_time_range,
)



from typing import Final

DEFAULT_READ_EC_PRODUCT_HEADER: Final[bool] = False
DEFAULT_READ_EC_PRODUCT_MODIFY: Final[bool] = True
DEFAULT_READ_EC_PRODUCT_META: Final[bool] = True
DEFAULT_READ_EC_PRODUCT_ENSURE_NANS: Final[bool] = True



from typing import Literal, Sequence

import matplotlib.colors
import numpy as np
from matplotlib.colors import Colormap

from ..cmap import Cmap
from ._basic_mpl_shift_colormap import shift_mpl_colormap


def shift_cmap(
    cmap: str | Colormap | None,
    start: float = 0.0,
    midpoint: float = 0.5,
    stop: float = 1.0,
    name: str = "shifted_cmap",
) -> Cmap:
    """Create a colormap with its center point shifted to a specified value.

    This function is useful for data with asymmetric ranges (e.g., negative min and
    positive max) where you want the center of the colormap to align with a specific
    value like zero.

    Args:
        cmap (str | Colormap | None): Colormap to be modified
        start (float): Lower bound of the colormap range (value between 0 and `midpoint`). Defaults to 0.0.
        midpoint (float): New center point of the colormap (value between 0 and 1). Defaults to 0.5.
            For data ranging from vmin to vmax where you want the center at value v,
            set midpoint = 1 - vmax/(vmax + abs(vmin))
        stop (float): Upper bound of the colormap range (value between `midpoint` and 1). Defaults to 1.0.
        name (str): Name of the new colormap. Defaults to "shifted_cmap".

    Returns:
        Cmap: New colormap with shifted center
    """
    from ..colormap import get_cmap

    cmap = get_cmap(cmap)
    cmap = shift_mpl_colormap(
        cmap,
        start=start,
        midpoint=midpoint,
        stop=stop,
        name=name,
    )
    cmap = get_cmap(cmap)

    return cmap



import os
from logging import Logger

from ...config import ECKConfig, get_default_config_filepath, read_config


def parse_path_to_config(
    path_to_config: str | None,
    logger: Logger | None = None,
) -> ECKConfig:
    if not isinstance(path_to_config, str):
        path_to_config = get_default_config_filepath()

    if not os.path.exists(path_to_config):
        raise FileNotFoundError(f"No config file found at <{path_to_config}>.")

    config = read_config(path_to_config)

    return config



import warnings
from numbers import Number
from typing import Any, Literal

import numpy as np
from numpy.typing import ArrayLike, DTypeLike, NDArray

from .debug import get_calling_function_name


def circular_nanmean(degrees: ArrayLike, axis: int | None = None) -> float:
    """
    Compute the circular mean of angles in degrees, ignoring NaNs.
    """
    radians = np.deg2rad(degrees)
    sin_sum = np.nanmean(np.sin(radians), axis=axis)
    cos_sum = np.nanmean(np.cos(radians), axis=axis)
    mean_angle = np.arctan2(sin_sum, cos_sum)
    return np.rad2deg(mean_angle)


def wrap_to_interval(a: ArrayLike, min: float, max: float) -> NDArray:
    """
    Wrap values in `a` to the interval [`min`, `max`).
    """
    a = np.array(a)
    interval = max - min
    return (a - min) % interval + min


def isascending(
    a: ArrayLike,
    raise_error: bool = False,
    result_constant: bool = True,
) -> bool:
    """
    Check whether a sequence is initially ascending.

    Parameters:
        lats (ArrayLike): Input sequence (e.g., `list`, `numpy.array`, etc.).
        raise_error (bool, optional): If True, raises ValueError if the sequence is too short (< 2). Defaults to False.
        result_constant (bool, optional): If True, a constant sequence counts as acending. Defaults to True.

    Returns:
        is_ascending (bool): True if the sequence is ascending, False otherwise.

    Raises:
        ValueError: If given `mode` is invalid.
    """
    _a: NDArray = np.array(a)
    _a = _a[~np.isnan(_a)]

    if len(_a) < 2:
        if raise_error:
            raise ValueError(
                f"too few latitude values ({len(_a)}) but at least 2 are needed."
            )
        return False
    diff = np.diff(_a)
    for d in diff:
        if d > 0:
            return True
        elif d < 0:
            return False
    return result_constant


def ismonotonic(
    a: ArrayLike,
    strict: bool = False,
    mode: Literal["any", "increasing", "decreasing"] = "any",
    raise_error: bool = False,
    ignore_nans: bool = True,
):
    """
    Check whether a sequence is monotonic.

    Parameters:
        a (ArrayLike): Input sequence (e.g., `list`, `numpy.array`, etc.).
        strict (bool, optional): If True, checks for strictly increasing or decreasing sequences.
            If False, allows equal adjacent elements. Defaults to False.
        mode (Literal['any', 'increasing', 'decreasing'], optional): Direction of monotonicity to check. Defaults to 'any'.
            - 'any': Checks if the sequence is either increasing or decreasing,
                     depending on the initial difference of the first two elements.
            - 'increasing': Checks only for increasing order.
            - 'decreasing': Checks only for decreasing order.
        raise_error (bool): If True, raises ValueError if the sequence is not monotonic.

    Returns:
        is_monotonic (bool): True if the sequence is monotonic according to the specified parameters, False otherwise.

    Raises:
        ValueError: If given `mode` is invalid.
    """
    a = np.asarray(a)
    if ignore_nans:
        a = a[~np.isnan(a)]

    signs = np.sign(np.diff(a))

    correct_signs = []

    if not strict:
        correct_signs.append(0)

    if mode == "any":
        i: int = 0
        while i < len(signs) - 1 and signs[i] == 0:
            i = i + 1

        if signs[i] != 0:
            correct_signs.append(signs[i])
    elif mode == "increasing":
        correct_signs.append(1)
    elif mode == "decreasing":
        correct_signs.append(-1)
    else:
        raise ValueError(
            f"invalid `mode` ('{mode}') given, but expecting 'any', 'increasing' or 'decreasing'"
        )

    is_monotonic = all([s in correct_signs for s in signs])

    if raise_error and not is_monotonic:
        raise TypeError(
            f"sequence must be monotonic but it is not (strict={strict}, mode='{mode}')"
        )

    return is_monotonic


def isndarray(input: Any, dtype: DTypeLike | None = None, raise_error: bool = False):
    """
    Returns True if `input` has type `numpy.ndarray` and also checks if `dtype` is lower/equal
    in type hierarchy if given (i.e. returns True if `input.dtype` is subtype of `dtype`).
    """
    if dtype is None:
        is_ndarray = isinstance(input, np.ndarray)
    else:
        is_ndarray = isinstance(input, np.ndarray) and np.issubdtype(input.dtype, dtype)

    if raise_error and not is_ndarray:
        dtype_str = "Any" if dtype is None else str(dtype)
        raise TypeError(
            f"expected type ndarray[{dtype_str}] for `input` but got {type(input).__name__}[{type(input[0]).__name__}]"
        )

    return is_ndarray


def lookup_value_by_number(n: float, numbers: NDArray, values: NDArray) -> Any:
    """
    Returns the value corresponding to the number closest to a given number, using interpolation.

    Parameters:
        n (float): A single number to look up.
        numbers (NDArray): A series of of monotonically increasing numbers.
        values (NDArray[Any]): A series of values corresponding to each number in `numbers`.

    Returns:
        v (Any): The value from `values` that corresponds to the closest number in `numbers` to `n`.

    Raises:
        ValueError: If `numbers` and `values` have different lengths.
    """
    if n is None:
        raise ValueError(f"{lookup_value_by_number.__name__}() missing `n`")
    if numbers is None:
        raise ValueError(f"{lookup_value_by_number.__name__}() missing `numbers`")
    if values is None:
        raise ValueError(f"{lookup_value_by_number.__name__}() missing `values`")

    n = float(n)
    numbers = np.asarray(numbers)
    values = np.asarray(values)

    if numbers.shape[0] == 0:
        raise ValueError(
            f"{lookup_value_by_number.__name__}() `numbers` is empty but needs at least on element"
        )
    if values.shape[0] != numbers.shape[0]:
        raise ValueError(
            f"{lookup_value_by_number.__name__}() First shapes must be the same for `values` ({values.shape[0]}) and `numbers` ({numbers.shape[0]})"
        )

    idx0 = int(np.searchsorted(numbers, n).astype(int) - 1)
    idx1 = int(np.searchsorted(numbers, n).astype(int))

    idx0 = int(np.min([len(numbers) - 1, np.max([0, idx0])]))
    idx1 = int(np.min([len(numbers) - 1, np.max([0, idx1])]))

    if numbers[idx0] > n:
        idx0 = idx1

    total_diff = numbers[idx1] - numbers[idx0]

    diff = n - numbers[idx0]

    if total_diff == 0:
        frac = 0
    else:
        frac = diff / total_diff

    total_amount = values[idx1] - values[idx0]
    v = values[idx0] + total_amount * frac

    return v


def get_number_range(
    start: float, end: float, freq: float | None = None, periods: int | None = None
) -> NDArray[np.floating | np.integer]:
    """
    Generates a sequence of numbers based on frequency or number of periods.

    Parameters:
        freq (float, optional): A number defining the frequency of sampled values in the sequence.
        periods (int, optional): A number of defining the number of evenly spaced values to sample.

    Returns:
        number_range (np.ndarray[np.floating | np.integer]): A sequence of numbers,
            either sampled by frequency or evenly spaced n times.
    """
    if freq is None and periods is None:
        raise TypeError(
            f"{get_number_range.__name__}() missing 1 required argument: 'freq' or 'periods'"
        )
    elif isinstance(freq, float) and isinstance(periods, int):
        raise TypeError(
            f"{get_number_range.__name__}() expected 1 out of the 2 required arguments 'freq' and 'periods' but got both"
        )
    elif isinstance(freq, float):
        mod = start % freq
        s = start - mod
        if mod != 0.0:
            s += freq
        mod = end % freq
        e = end - mod
        result = np.arange(s, e + freq, freq)
        return np.array(result)
    elif isinstance(periods, int):
        result = np.linspace(start, end, periods)
        return np.array(result)
    else:
        raise RuntimeError(f"{get_number_range.__name__}() implementation error")


def normalize(
    values: list[Number] | np.ndarray,
    vmin: float = 0,
    vmax: float = 1,
) -> np.ndarray:
    """
    Normalizes a list or array of numbers to a specified range [vmin, vmax], preserving NaNs.

    The input is linearly scaled such that the minimum non-NaN value maps to `vmin`
    and the maximum to `vmax`. NaN values are preserved in their original positions.

    Args:
        values (list[Number] | np.ndarray): A sequence of numeric values, possibly containing NaNs.
        vmin (float): The minimum value of the normalized output range. Defaults to 0.
        vmax (float): The maximum value of the normalized output range. Defaults to 1.

    Returns:
        A `numpy.array` of the same shape as `values`, with values scaled to [vmin, vmax]
        and NaNs preserved.

    Raises:
        ValueError: If `vmin` is equal (i.e. zero output range) or greater than `vmax`.
    """
    if vmin >= vmax:
        raise ValueError(f"vmin ({vmin}) must be smaller than vmax ({vmax})")

    a_old = np.asarray(values, dtype=float)
    vmin_old = np.nanmin(a_old)
    vmax_old = np.nanmax(a_old)

    if np.isnan(vmin_old) or vmin_old == vmax_old:
        a_new = np.full_like(a_old, np.nan)
    else:
        a_new = (a_old - vmin_old) / (vmax_old - vmin_old)

    # Scale
    a_new = a_new * (vmax - vmin)

    # Shift
    a_new = a_new + vmin

    return a_new


def all_same(a: ArrayLike) -> bool:
    """
    Check if all elements in the input array are the same.

    Args:
        a (ArrayLike): Input array or array-like object to check.

    Returns:
        bool: True if all elements in the array are the same, False otherwise.
    """
    a = np.asarray(a)
    return bool(np.all(a == a[0]))


def pad_true_sequence(
    a: NDArray[np.bool_],
    n: int,
) -> NDArray[np.bool_]:
    """Pads all sequences of True values occuring in an array with n True values before and after the sequence (while keeping the original size)."""
    idx = np.flatnonzero(a)
    if idx.size == 0:
        return a.copy()
    mask: NDArray[np.bool_] = np.full(a.shape, False, dtype=bool)
    start = max(0, idx[0] - n)
    end = min(len(a), idx[-1] + n + 1)
    mask[start:end] = True
    return mask


def get_most_freq_int(a: ArrayLike):
    a = np.asarray(a)
    min_val = a.min()
    if min_val < 0:
        shifted = a - min_val
        return np.bincount(shifted).argmax() + min_val
    else:
        return np.bincount(a).argmax()


def coarsen_mean(
    a: ArrayLike,
    n: int,
    axis: int = 0,
    is_bin: bool = False,
) -> NDArray:
    """
    Downsamples a array by averaging every n adjacient elements together, discarding residual elements at the end.

    Args:
        a (ArrayLike): Input array or array-like object to downsample.
        n (int): Number of elements to be averaged together.
        axis (int): The axis along which the array `a` will be downsampled.

    Returns:
        np.ndarray: The downsampled array.
    """
    a = np.asarray(a)
    a = np.moveaxis(a, axis, 0)

    # Discard residual data points
    trimmed_len = (a.shape[0] // n) * n
    trimmed = a[:trimmed_len]
    reshaped = trimmed.reshape(-1, n, *a.shape[1:])

    # Average
    is_datetime = np.issubdtype(a.dtype, np.datetime64)
    averaged: NDArray
    if is_datetime:
        averaged = reshaped.astype("datetime64[ns]").astype("int64").mean(axis=1)
        averaged = averaged.astype("datetime64[ns]")
    elif is_bin:
        if len(a.shape) == 1:
            averaged = np.apply_along_axis(get_most_freq_int, 0, reshaped)
        elif len(a.shape) == 2:
            averaged = np.array(
                [np.apply_along_axis(get_most_freq_int, 0, x) for x in reshaped]
            )
    else:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            averaged = np.nanmean(reshaped, axis=1)

    return np.moveaxis(averaged, 0, axis)


def flatten_array(sequence: ArrayLike) -> NDArray:
    """Flatten a nested sequence of array-likes into a 1D numpy.array.

    Args:
        sequence (ArrayLike): Sequence of array-like objects (may contain lists, tuples, arrays, or non-iterable elements).

    Returns:
        np.ndarray: Flattened 1D array.
    """
    if isinstance(sequence, np.ndarray):
        return sequence.flatten()

    flattened_sequence = []

    def _ensure_list(x):
        if isinstance(x, list):
            return x.copy()
        return [x].copy()

    stack = _ensure_list(sequence)  # type: ignore

    while stack:
        item = stack.pop(0)
        if isinstance(item, (list, tuple, np.ndarray)):
            stack = list(item) + stack
        else:
            flattened_sequence.append(item)

    return np.array(flattened_sequence)


def clamp(a: ArrayLike, min: float, max: float) -> NDArray:
    """Limits given values to a range between a minimum and maximum value.

    Args:
        a (ArrayLike): Input array or array-like object to be clamped.
        min (float): Minimum limit.
        max (float): Maximum limit.

    Returns:
        NDArray: Clampled array.
    """
    if np.isnan(max):
        max = np.nanmax(a)
    if np.isnan(min):
        min = np.nanmin(a)
    return np.maximum(np.minimum(np.asarray(a), max), min)



"""Ensures that all earthcare data folders are moved correctly in the subfolder structure given by earthcarekit"""

import os
import re
import shutil
from logging import Logger
from pathlib import Path

import pandas as pd

from ..utils.config import ECKConfig
from ..utils.time import time_to_string
from ._eo_product import get_local_product_dirpath


def delete_dir(
    dirpath: str,
    logger: Logger | None = None,
) -> None:
    """Delete a directory and all its contents."""
    if os.path.exists(dirpath):
        shutil.rmtree(dirpath)
        if logger:
            logger.info(f"Deleted directory: <{dirpath}>")
    elif logger:
        logger.info(f"Directory not found: <{dirpath}>")


def organize_data(
    config: ECKConfig,
    dry_run: bool = False,
    create_log_file: bool = False,
    logger: Logger | None = None,
) -> list[dict[str, str]]:
    root_directory = config.path_to_data

    pattern: str = (
        r"^ECA_[EJ][XNO][A-Z]{2}_.........._\d{8}T\d{6}Z_\d{8}T\d{6}Z_\d{5}[ABCDEFGH](\.h5|\.HDR)$"
    )
    time_now: str = time_to_string(pd.Timestamp.now(), format="%Y%m%dT%H%M%S")
    log_filepath: str = f"./organize_data_{time_now}.csv"

    moves_performed: list[dict[str, str]] = []

    for current_dir, subdirs, files in os.walk(root_directory, topdown=False):
        for file in files[:]:
            folder_path: Path = Path(os.path.join(current_dir, file))

            match = re.match(pattern, file)
            if match:
                product_folder = file.split(".")[0]

                target_base: Path = Path(
                    get_local_product_dirpath(
                        root_directory,
                        product_folder,
                        create_subdirs=True,
                        config=config,
                    )
                )
                target_dirpath: Path = Path(os.path.join(target_base, product_folder))
                target_path: Path = Path(
                    os.path.join(target_base, product_folder, file)
                )

                if folder_path == target_path:
                    continue

                move_info = {
                    "source": str(folder_path),
                    "target": str(target_path),
                    "status": "dry_run",
                    "error": "",
                }

                if dry_run:
                    if logger:
                        logger.info(
                            f"[DRY RUN] Would move: <{str(folder_path)}> -> <{str(target_path)}>"
                        )
                        moves_performed.append(move_info)
                else:
                    try:
                        target_dirpath.mkdir(parents=True, exist_ok=True)

                        shutil.move(str(folder_path), str(target_path))
                        if logger:
                            logger.info(
                                f"Moved: <{str(folder_path)}> -> <{str(target_path)}>"
                            )
                        move_info["status"] = "success"
                        moves_performed.append(move_info)

                        # subdirs.remove(subdir)
                    except Exception as e:
                        if logger:
                            logger.info(f"Error moving <{str(folder_path)}>: {e}")

                        if str(target_path) in str(folder_path):
                            delete_dir(str(folder_path), logger=logger)

                        move_info["status"] = "error"
                        move_info["error"] = str(e)
                        moves_performed.append(move_info)

        # Remove empty folders
        with os.scandir(current_dir) as entries:
            has_subdirs = any(entry.is_dir() for entry in entries)
            has_files = len(os.listdir(current_dir)) > 0
            if not has_files and not has_subdirs:
                os.rmdir(current_dir)
                if logger:
                    logger.info(
                        f"Removed empty folder: <{os.path.abspath(current_dir)}>"
                    )

    df = pd.DataFrame(moves_performed)
    if df.size > 0:
        if create_log_file:
            df.to_csv(log_filepath)
            if logger:
                logger.info(f"Saved log to <{os.path.abspath(log_filepath)}>")
    else:
        if logger:
            logger.info("No moves were performed")

    return moves_performed



from logging import Logger
from typing import Literal, Sequence

import numpy as np
import pandas as pd
import xarray as xr
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ....utils.constants import CM_AS_INCH, TIME_VAR
from ....utils.time import TimedeltaLike, TimeRangeLike
from ....utils.typing import DistanceRangeLike
from ....utils.xarray_utils import filter_radius, filter_time
from ...figure import (
    CurtainFigure,
    ECKFigure,
    FigureType,
    MapFigure,
    create_multi_figure_layout,
)
from .._cli import print_progress
from .._quicklook_results import QuicklookFigure
from ..set_default_height_range import set_none_height_range_to_default


def ecquicklook_atc(
    ds: xr.Dataset,
    vars: list[str] | None = None,
    show_maps: bool = True,
    show_zoom: bool = False,
    show_profile: bool = False,
    site: str | None = None,
    radius_km: float = 100.0,
    time_range: TimeRangeLike | None = None,
    height_range: DistanceRangeLike | None = (0, 30e3),
    ds_tropopause: xr.Dataset | None = None,
    ds_elevation: xr.Dataset | None = None,
    ds_temperature: xr.Dataset | None = None,
    logger: Logger | None = None,
    log_msg_prefix: str = "",
    selection_max_time_margin: TimedeltaLike | Sequence[TimedeltaLike] | None = None,
    mode: Literal["fast", "exact"] = "fast",
) -> QuicklookFigure:
    _stime: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")

    height_range = set_none_height_range_to_default(height_range, 0, 30e3)

    if ds_elevation is None:
        ds_elevation = ds

    if vars is None:
        vars = [
            "simple_classification",
            "classification_low_resolution",
            "classification_medium_resolution",
            "classification",
        ]

    wspaces: list[float] = []

    if show_maps:
        map_rows = [FigureType.MAP_1_ROW, FigureType.MAP_1_ROW]
        wspaces.append(3.5 * CM_AS_INCH)
    else:
        map_rows = None

    main_rows = [FigureType.CURTAIN for _ in vars]

    if show_zoom:
        zoom_rows = [FigureType.CURTAIN_ZOOMED for _ in vars]
        wspaces.append(6.5 * CM_AS_INCH)
    else:
        zoom_rows = None

    profile_rows = None

    if logger:
        print_progress(f"layout", log_msg_prefix=log_msg_prefix, logger=logger)

    output = create_multi_figure_layout(
        map_rows=map_rows,
        rows=main_rows,
        zoom_rows=zoom_rows,
        profile_rows=profile_rows,
        wspace=wspaces,
    )
    fig = output.fig
    axs_map = output.axs_map
    axs_main = output.axs
    axs_zoom = output.axs_zoom
    axs_profile = output.axs_profile

    map_figs: list[ECKFigure] = []
    main_figs: list[ECKFigure] = []
    zoom_figs: list[ECKFigure] = []

    if show_maps:
        if logger:
            print_progress(f"map globe", log_msg_prefix=log_msg_prefix, logger=logger)
        mf = MapFigure(ax=axs_map[0])
        mf = mf.ecplot(
            ds,
            site=site,
            radius_km=radius_km,
            time_range=time_range,
            selection_max_time_margin=selection_max_time_margin,
        )
        map_figs.append(mf)

        if logger:
            print_progress(f"map zoomed", log_msg_prefix=log_msg_prefix, logger=logger)
        mf = MapFigure(
            ax=axs_map[1],
            style="blue_marble",
            coastlines_resolution="50m",
            show_night_shade=False,
            show_right_labels=False,
            show_top_labels=False,
        )
        mf = mf.ecplot(
            ds,
            site=site,
            radius_km=radius_km,
            time_range=time_range,
            view="overpass",
            selection_max_time_margin=selection_max_time_margin,
        )
        map_figs.append(mf)

    for i, var in enumerate(vars):
        if logger:
            print_progress(
                f"curtain: {var=}", log_msg_prefix=log_msg_prefix, logger=logger
            )
        cf = CurtainFigure(ax=axs_main[i], mode=mode)
        cf = cf.ecplot(
            ds,
            var,
            site=site,
            radius_km=radius_km,
            selection_time_range=time_range,
            height_range=height_range,
            selection_max_time_margin=selection_max_time_margin,
        )
        if ds_tropopause:
            if logger:
                print_progress(
                    f"tropopause", log_msg_prefix=log_msg_prefix, logger=logger
                )
            cf = cf.ecplot_tropopause(ds_tropopause)
        if ds_temperature:
            if logger:
                print_progress(
                    f"temperature", log_msg_prefix=log_msg_prefix, logger=logger
                )
            cf = cf.ecplot_temperature(ds_temperature)

        main_figs.append(cf)

    if show_zoom or show_profile:
        _ds = ds.copy()
        if site:
            _ds = filter_radius(_ds, radius_km=radius_km, site=site)
        elif time_range:
            _ds = filter_time(_ds, time_range)

        if show_zoom:
            _time_range = _ds[TIME_VAR].values[[0, -1]]
            for i, var in enumerate(vars):
                if logger:
                    print_progress(
                        f"{log_msg_prefix}curtain zoomed: {var=} ...",
                        log_msg_prefix=log_msg_prefix,
                        logger=logger,
                    )
                cf = CurtainFigure(
                    ax=axs_zoom[i],
                    num_ticks=4,
                    show_height_left=False,
                    show_height_right=True,
                    mode=mode,
                )
                cf = cf.ecplot(
                    ds,
                    var,
                    colorbar=False,
                    time_range=_time_range,
                    height_range=height_range,
                )
                if ds_tropopause:
                    if logger:
                        print_progress(
                            f"tropopause zoomed",
                            log_msg_prefix=log_msg_prefix,
                            logger=logger,
                        )
                    cf = cf.ecplot_tropopause(ds_tropopause)
                if ds_temperature:
                    if logger:
                        print_progress(
                            f"temperature zoomed",
                            log_msg_prefix=log_msg_prefix,
                            logger=logger,
                        )
                    cf = cf.ecplot_temperature(ds_temperature)

                zoom_figs.append(cf)

    subfigs: list[list[ECKFigure]] = [map_figs, main_figs, zoom_figs]

    _etime: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")
    _dtime: str = str(pd.Timestamp(_etime) - pd.Timestamp(_stime)).split()[-1]
    if logger:
        print_progress(
            f"Plot created (time taken {_dtime}).",
            is_last=True,
            log_msg_prefix=log_msg_prefix,
            logger=logger,
        )

    return QuicklookFigure(fig, subfigs)



import numpy as np
import xarray as xr
from scipy.spatial import cKDTree  # type: ignore

from ...constants import (
    ALONG_TRACK_DIM,
    HEIGHT_VAR,
    TIME_VAR,
    TRACK_LAT_VAR,
    TRACK_LON_VAR,
    VERTICAL_DIM,
)
from ...geo import sequence_geo_to_ecef
from ...xarray_utils import remove_dims
from ._generic import read_product
from .auxiliary.aux_met_1d import read_product_xmet


def rebin_xmet_to_vertical_track(
    ds_xmet: xr.Dataset | str,
    ds_vert: xr.Dataset | str,
    vars: list[str] | None = None,
    k: int = 4,
    eps: float = 1e-12,
    lat_var: str = TRACK_LAT_VAR,
    lon_var: str = TRACK_LON_VAR,
    time_var: str = TIME_VAR,
    height_var: str = HEIGHT_VAR,
    along_track_dim: str = ALONG_TRACK_DIM,
    height_dim: str = VERTICAL_DIM,
    xmet_lat_var: str = "latitude",
    xmet_lon_var: str = "longitude",
    xmet_height_var: str = "geometrical_height",
    xmet_height_dim: str = "height",
    xmet_horizontal_grid_dim: str = "horizontal_grid",
) -> xr.Dataset:
    """
    Rebins variables from an AUX_MET_1D (XMET) dataset onto the vertical curtain track of given by another dataset (e.g. ATL_EBD_2A).

    This function interpolates selected variables from `ds_xmet` onto a EarthCARE
    vertical track given in `ds_vert`, using quick horizontal kd-tree nearest-neighbor search with `scipy.spatial.cKDTree` followed
    by averaging the `k`-nearest vertical XMET profiles using inverse distance weighting. The resulting
    profiles are then interpolated in the vertical to match the height resolution of `ds_vert`.

    Args:
        ds_xmet (xr.Dataset | str): The source XMET dataset from which vertical curtain along track will be interpolated.
        ds_vert (xr.Dataset | str): The target dataset containing the vertical curtain track.
        vars (list[str] | None, optional): List of variable names from `ds_xmet` to rebin.
            If None, all data variables are considered.
        k (int, optional): Number of nearest horizontal neighbors to include in the kd-tree search.
            Defaults to 4.
        eps (float, optional): Numerical threshold to avoid division by zero in distance calculations during the kd-tree search.
            Defaults to 1e-12.
        lat_var (str, optional): Name of the latitude variable in `ds_vert`.
            Defaults to TRACK_LAT_VAR.
        lon_var (str, optional): Name of the longitude variable in `ds_vert`.
            Defaults to TRACK_LON_VAR.
        time_var (str, optional): Name of the time variable in `ds_vert`.
            Defaults to TIME_VAR.
        height_var (str, optional): Name of the height variable in `ds_vert`.
            Defaults to HEIGHT_VAR.
        along_track_dim (str, optional): Name of the along-track dimension in `ds_vert`.
            Defaults to ALONG_TRACK_DIM.
        height_dim (str, optional): Name of the vertical or height dimension in `ds_vert`.
            Defaults to VERTICAL_DIM.
        xmet_lat_var (str, optional): Name of the latitude variable in `ds_xmet`.
            Defaults to "latitude".
        xmet_lon_var (str, optional): Name of the longitude variable in `ds_xmet`.
            Defaults to "longitude".
        xmet_height_var (str, optional): Name of the height variable in `ds_xmet`.
            Defaults to "geometrical_height".
        xmet_height_dim (str, optional): Name of the vertical dimension in `ds_xmet`.
            Defaults to "height".
        xmet_horizontal_grid_dim (str, optional): Name of the horizontal grid dimension in `ds_xmet`.
            Defaults to "horizontal_grid".

    Returns:
        xr.Dataset: A new dataset containing the selected XMET variables interpolated to the grid of the
            vertical curtain given in `ds_vert`. This new dataset has the same along-track and vertical
            dimensions as `ds_vert`.

    Raises:
        KeyError: If any specified variable or coordinate name is not found in `ds_xmet`.
    """

    def _read_xmet() -> xr.Dataset:
        if isinstance(ds_xmet, str):
            return read_product_xmet(ds_xmet)
        return ds_xmet

    def _read_vert() -> xr.Dataset:
        if isinstance(ds_vert, str):
            return read_product(ds_vert)
        return ds_vert

    with (
        _read_xmet() as ds_xmet,
        _read_vert() as ds_vert,
    ):

        if vars is None:
            vars = [str(v) for v in ds_xmet.variables]
        else:
            for var in vars:
                if var not in ds_xmet.variables:
                    present_vars = [str(v) for v in ds_xmet.variables]
                    raise KeyError(
                        f"""X-MET dataset does not contain variable "{var}". Present variables are: {", ".join(present_vars)}"""
                    )

        new_ds_xmet = ds_xmet.copy().swap_dims({xmet_height_dim: "tmp_xmet_height"})
        new_ds_xmet[time_var] = ds_vert[time_var].copy()
        new_ds_xmet[height_var] = ds_vert[height_var].copy()

        hgrid_lat = ds_xmet[xmet_lat_var].values.flatten()
        hgrid_lon = ds_xmet[xmet_lon_var].values.flatten()
        hgrid_alt = ds_xmet[xmet_height_var].values
        hgrid_coords = sequence_geo_to_ecef(hgrid_lat, hgrid_lon)

        track_lat = ds_vert[lat_var].values
        track_lon = ds_vert[lon_var].values
        track_alt = ds_vert[height_var].values
        track_coords = sequence_geo_to_ecef(track_lat, track_lon)

        tree = cKDTree(hgrid_coords)
        dists, idxs = tree.query(track_coords, k=k)

        # Inverse distance weighting
        if k > 1:
            weights = 1.0 / (dists + eps)
            weights /= np.sum(weights, axis=1, keepdims=True)
            height = np.einsum("ij,ijh->ih", weights, hgrid_alt[idxs])
        else:
            weights = np.ones(idxs.shape)
            height = hgrid_alt[idxs]

        dims: str | tuple[str, str]
        for var in vars:
            values = ds_xmet[var].values
            if len(values.shape) == 0:
                continue

            if len(values.shape) == 1:
                dims = along_track_dim

                if k > 1:
                    result = np.sum(values[idxs] * weights, axis=1)
                    new_values = result
                else:
                    new_values = values[idxs]
            else:
                dims = (along_track_dim, height_dim)

                if k > 1:
                    result = np.einsum("ij,ijh->ih", weights, values[idxs])
                else:
                    result = values[idxs]

                new_values = np.empty(track_alt.shape)
                new_values[:] = np.nan

                for i in np.arange(track_alt.shape[0]):
                    _new_values = np.interp(
                        track_alt[i],
                        height[i],
                        result[i],
                    )

                    # FIXME: Is this still needed?
                    # _new_values = interp(track_alt[i])
                    # Fill nans
                    # _new_values[np.isnan(_new_values) & (track_alt[i] < height[i, 0])] = result[i, 0]
                    # _new_values[np.isnan(_new_values) & (track_alt[i] > height[i, -1])] = result[i, -1]

                    new_values[i] = _new_values

            new_var = f"{var}"
            new_ds_xmet[new_var] = (dims, new_values)
            new_ds_xmet[new_var].attrs = ds_xmet[var].attrs

        new_ds_xmet = remove_dims(
            new_ds_xmet, [xmet_horizontal_grid_dim, xmet_height_dim]
        )

        return new_ds_xmet



import numpy as np
import xarray as xr

from ....constants import (
    ACROSS_TRACK_DIM,
    ALONG_TRACK_DIM,
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
    SWATH_LAT_VAR,
    SWATH_LON_VAR,
)
from ....swath_data.across_track_distance import (
    add_across_track_distance,
    add_nadir_track,
    drop_samples_with_missing_geo_data_along_track,
    get_nadir_index,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import rename_common_dims_and_vars, rename_var_info
from ..file_info import FileAgency
from ..science_group import read_science_data


def add_isccp_cloud_type(
    ds: xr.Dataset,
    new_var: str = "isccp_cloud_type",
    cot_var: str = "cloud_optical_thickness",
    cth_var: str = "cloud_top_height",
    along_track_dim: str = ALONG_TRACK_DIM,
    across_track_dim: str = ACROSS_TRACK_DIM,
) -> xr.Dataset:
    """
    Adds a variable to the dataset containing ISCCP cloud types calculated from cloud optical thickness (COT)
    and cloud top height (CTH).

    Args:
        ds (xr.Dataset): A MSI_COP_2A dataset.
        new_var (str, optional): Name of the new ISCCP cloud type variable. Defaults to "isccp_cloud_type".
        cot_var (str, optional): Name of the COT variable in `ds`. Defaults to "cloud_optical_thickness".
        cth_var (str, optional): Name of the CTH variable in `ds`. Defaults to "cloud_top_height".
        along_track_dim (str, optional): Name of the along-track dimension in `ds`. Defaults to ALONG_TRACK_DIM.
        across_track_dim (str, optional): Name of the across-track dimension in `ds`. Defaults to ACROSS_TRACK_DIM.

    Returns:
        xr.Dataset: The input dataset with added ISCCP cloud type variable.

    References:
        International Satellite Cloud Climatology Project (ISCCP), "ISCCP Definition of Cloud Types."
            https://isccp.giss.nasa.gov/cloudtypes.html (accessed 2025-09-25)
    """
    cot = ds[cot_var].values
    cth = ds[cth_var].values

    cu = np.where((cth >= 100) & (cth < 3200) & (cot >= 0.01) & (cot < 3.6))
    ac = np.where((cth >= 3200) & (cth < 6500) & (cot >= 0.01) & (cot < 3.6))
    ci = np.where((cth >= 6500) & (cth < 19300) & (cot >= 0.01) & (cot < 3.6))
    sc = np.where((cth >= 100) & (cth < 3200) & (cot >= 3.6) & (cot < 23))
    asc = np.where((cth >= 3200) & (cth < 6500) & (cot >= 3.6) & (cot < 23))
    cs = np.where((cth >= 6500) & (cth < 19300) & (cot >= 3.6) & (cot < 23))
    st = np.where((cth >= 100) & (cth < 3200) & (cot >= 23))
    ns = np.where((cth >= 3200) & (cth < 6500) & (cot >= 23))
    cb = np.where((cth >= 6500) & (cth < 19300) & (cot >= 23))
    clear = np.where((cot < 0.01) & (cot >= 0))

    cloud_type = np.empty(shape=cot.shape, dtype=int)
    cloud_type[:, :] = -127

    cloud_type[cu] = 1
    cloud_type[ac] = 2
    cloud_type[ci] = 3
    cloud_type[sc] = 4
    cloud_type[asc] = 5
    cloud_type[cs] = 6
    cloud_type[st] = 7
    cloud_type[ns] = 8
    cloud_type[cb] = 9
    cloud_type[clear] = 0

    da = xr.DataArray(
        cloud_type,
        dims=(along_track_dim, across_track_dim),
        name=new_var,
        attrs={
            "units": "",
            "long_name": "ISCCP cloud type calculated from M-COP",
            "definition": "0: Clear, 1: Cumulus, 2: Altocumulus, 3: Cirrus, 4: Stratocumulus, 5: Altostratus, 6: Cirrostratus, 7: Stratus, 8: Nimbostratus, 9: Deep convection, -127: Not determined",
            "earthcarekit": "Added by earthcarekit",
        },
    )
    ds[new_var] = da

    return ds


def read_product_mcop(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> xr.Dataset:
    """Opens MSI_COP_2A file as a `xarray.Dataset`."""
    ds = read_science_data(
        filepath,
        agency=FileAgency.ESA,
        ensure_nans=ensure_nans,
        **kwargs,
    )

    if not modify:
        return ds

    ds = drop_samples_with_missing_geo_data_along_track(
        ds=ds,
        swath_lat_var="latitude",
        along_track_dim="along_track",
        across_track_dim="across_track",
    )

    nadir_idx = get_nadir_index(ds, nadir_idx=266)
    ds = ds.rename({"latitude": SWATH_LAT_VAR})
    ds = ds.rename({"longitude": SWATH_LON_VAR})
    ds = add_nadir_track(
        ds,
        nadir_idx,
        swath_lat_var=SWATH_LAT_VAR,
        swath_lon_var=SWATH_LON_VAR,
        along_track_dim="along_track",
        across_track_dim="across_track",
        nadir_lat_var="latitude",
        nadir_lon_var="longitude",
    )
    ds = add_across_track_distance(
        ds,
        nadir_idx,
        swath_lat_var=SWATH_LAT_VAR,
        swath_lon_var=SWATH_LON_VAR,
    )

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        across_track_dim="across_track",
        track_lat_var="latitude",
        track_lon_var="longitude",
        swath_lat_var=SWATH_LAT_VAR,
        swath_lon_var=SWATH_LON_VAR,
        time_var="time",
    )

    ds = add_isccp_cloud_type(ds)

    return ds



import os
from typing import overload

import xarray as xr

from ..header_group import read_header_data
from .file_info import FileInfoEnum


class FileLatency(FileInfoEnum):
    NEAR_REAL_TIME = "N"
    OFFLINE = "O"
    NOT_APPLICABLE = "X"

    @classmethod
    def from_input(cls, input: str | xr.Dataset) -> "FileLatency":
        """Infers the EarthCARE product latency indicator (i.e. N for Near-real time, O for Offline, X for not applicable) from a given name, file or dataset."""
        if isinstance(input, str):
            try:
                return cls[input.upper()]
            except AttributeError:
                pass
            except KeyError:
                pass
            try:
                return cls(input.upper())
            except ValueError:
                pass

        return get_file_latency(input)


def _get_file_latency_from_dataset(ds: xr.Dataset) -> FileLatency:
    try:
        return FileLatency(str(ds.File_Class.values)[1])
    except AttributeError as e:
        filepath = ds.encoding["source"]
        filename = os.path.basename(filepath)
        file_class = filename.split(".")[0].split("_")[1]
        return FileLatency(file_class[1])


@overload
def get_file_latency(product: str) -> FileLatency: ...
@overload
def get_file_latency(product: xr.Dataset) -> FileLatency: ...
def get_file_latency(product: str | xr.Dataset) -> FileLatency:
    if isinstance(product, str):
        with read_header_data(product) as ds:
            file_class = _get_file_latency_from_dataset(ds)
    elif isinstance(product, xr.Dataset):
        file_class = _get_file_latency_from_dataset(product)
    else:
        raise NotImplementedError()
    return file_class



import logging
from typing import Iterable, Literal

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib import font_manager
from matplotlib.axes import Axes
from matplotlib.colorbar import Colorbar
from matplotlib.colors import Colormap, LogNorm, Normalize
from matplotlib.dates import date2num
from matplotlib.figure import Figure, SubFigure
from matplotlib.offsetbox import AnchoredOffsetbox, AnchoredText
from matplotlib.text import Text
from numpy.typing import ArrayLike, NDArray

from ...utils.constants import *
from ...utils.constants import FIGURE_HEIGHT_SWATH, FIGURE_WIDTH_SWATH
from ...utils.swath_data import SwathData
from ...utils.swath_data.across_track_distance import get_nadir_index
from ...utils.time import (
    TimeRangeLike,
    TimestampLike,
    to_timestamp,
    validate_time_range,
)
from ...utils.typing import DistanceRangeLike, ValueRangeLike
from ..color import Cmap, Color, ColorLike, get_cmap
from ..save import save_plot
from ._ensure_updated_msi_rgb_if_required import ensure_updated_msi_rgb_if_required
from .along_track import AlongTrackAxisStyle, format_along_track_axis
from .annotation import add_text_product_info, format_var_label
from .colorbar import add_colorbar
from .defaults import get_default_cmap, get_default_norm
from .height_ticks import format_height_ticks

logger = logging.getLogger(__name__)


class SwathFigure:
    """TODO: documentation"""

    def __init__(
        self,
        ax: Axes | None = None,
        figsize: tuple[float, float] = (FIGURE_WIDTH_SWATH, FIGURE_HEIGHT_SWATH),
        dpi: int | None = None,
        title: str | None = None,
        ax_style_top: AlongTrackAxisStyle | str = "geo",
        ax_style_bottom: AlongTrackAxisStyle | str = "time",
        num_ticks: int = 10,
        colorbar_tick_scale: float | None = None,
        ax_style_y: Literal[
            "from_track_distance",
            "across_track_distance",
            "pixel",
        ] = "from_track_distance",
        fig_height_scale: float = 1.0,
        fig_width_scale: float = 1.0,
    ):
        figsize = (figsize[0] * fig_width_scale, figsize[1] * fig_height_scale)
        self.fig: Figure
        if isinstance(ax, Axes):
            tmp = ax.get_figure()
            if not isinstance(tmp, (Figure, SubFigure)):
                raise ValueError(f"Invalid Figure")
            self.fig = tmp  # type: ignore
            self.ax = ax
        else:
            self.fig = plt.figure(figsize=figsize, dpi=dpi)
            self.ax = self.fig.add_axes((0.0, 0.0, 1.0, 1.0))

        self.title = title
        if self.title:
            self.fig.suptitle(self.title)

        self.ax_top: Axes | None = None
        self.ax_right: Axes | None = None
        self.colorbar: Colorbar | None = None
        self.colorbar_tick_scale: float | None = colorbar_tick_scale
        self.selection_time_range: tuple[pd.Timestamp, pd.Timestamp] | None = None
        self.ax_style_top: AlongTrackAxisStyle = AlongTrackAxisStyle.from_input(
            ax_style_top
        )
        self.ax_style_bottom: AlongTrackAxisStyle = AlongTrackAxisStyle.from_input(
            ax_style_bottom
        )
        self.ax_style_y: Literal[
            "from_track_distance",
            "across_track_distance",
            "pixel",
        ] = ax_style_y

        self.info_text: AnchoredText | None = None
        self.info_text_loc: str = "upper right"
        self.num_ticks = num_ticks

    def _set_info_text_loc(self, info_text_loc: str | None) -> None:
        if isinstance(info_text_loc, str):
            self.info_text_loc = info_text_loc

    def plot(
        self,
        swath: SwathData | None = None,
        *,
        values: NDArray | None = None,
        time: NDArray | None = None,
        nadir_index: int | None = None,
        latitude: NDArray | None = None,
        longitude: NDArray | None = None,
        # Common args for wrappers
        value_range: ValueRangeLike | None = None,
        log_scale: bool | None = None,
        norm: Normalize | None = None,
        time_range: TimeRangeLike | None = None,
        from_track_range: DistanceRangeLike | None = None,
        label: str | None = None,
        units: str | None = None,
        cmap: str | Colormap | None = None,
        colorbar: bool = True,
        colorbar_ticks: ArrayLike | None = None,
        colorbar_tick_labels: ArrayLike | None = None,
        colorbar_position: str | Literal["left", "right", "top", "bottom"] = "right",
        colorbar_alignment: str | Literal["left", "center", "right"] = "center",
        colorbar_width: float = DEFAULT_COLORBAR_WIDTH,
        colorbar_spacing: float = 0.2,
        colorbar_length_ratio: float | str = "100%",
        colorbar_label_outside: bool = True,
        colorbar_ticks_outside: bool = True,
        colorbar_ticks_both: bool = False,
        selection_time_range: TimeRangeLike | None = None,
        selection_color: str | None = Color("ec:earthcare"),
        selection_linestyle: str | None = "dashed",
        selection_linewidth: float | int | None = 2.5,
        selection_highlight: bool = False,
        selection_highlight_inverted: bool = True,
        selection_highlight_color: str = Color("white"),
        selection_highlight_alpha: float = 0.5,
        ax_style_top: AlongTrackAxisStyle | str | None = None,
        ax_style_bottom: AlongTrackAxisStyle | str | None = None,
        ax_style_y: (
            Literal["from_track_distance", "across_track_distance", "pixel"] | None
        ) = None,
        show_nadir: bool = True,
        nadir_color: ColorLike | None = "red",
        nadir_linewidth: int | float = 1.5,
        label_length: int = 25,
        **kwargs,
    ) -> "SwathFigure":
        if isinstance(value_range, Iterable):
            if len(value_range) != 2:
                raise ValueError(
                    f"invalid `value_range`: {value_range}, expecting (vmin, vmax)"
                )
        else:
            value_range = (None, None)

        cmap = get_cmap(cmap)

        if isinstance(cmap, Cmap) and cmap.categorical == True:
            norm = cmap.norm
        elif isinstance(norm, Normalize):
            if log_scale == True and not isinstance(norm, LogNorm):
                norm = LogNorm(norm.vmin, norm.vmax)
            elif log_scale == False and isinstance(norm, LogNorm):
                norm = Normalize(norm.vmin, norm.vmax)
            if value_range[0] is not None:
                norm.vmin = value_range[0]  # type: ignore
            if value_range[1] is not None:
                norm.vmax = value_range[1]  # type: ignore
        else:
            if log_scale == True:
                norm = LogNorm(value_range[0], value_range[1])  # type: ignore
            else:
                norm = Normalize(value_range[0], value_range[1])  # type: ignore

        assert isinstance(norm, Normalize)
        value_range = (norm.vmin, norm.vmax)

        if isinstance(swath, SwathData):
            values = swath.values
            time = swath.time
            nadir_index = swath.nadir_index
            latitude = swath.latitude
            longitude = swath.longitude
            label = swath.label
            units = swath.units
        elif (
            values is None
            or time is None
            or nadir_index is None
            or latitude is None
            or longitude is None
        ):
            raise ValueError(
                "Missing required arguments. Provide either a `SwathData` or all of `values`, `time`, `nadir_index`, `latitude` and `longitude`"
            )

        values = np.asarray(values)
        time = np.asarray(time)
        latitude = np.asarray(latitude)
        longitude = np.asarray(longitude)

        swath_data = SwathData(
            values=values,
            time=time,
            latitude=latitude,
            longitude=longitude,
            nadir_index=nadir_index,
            label=label,
            units=units,
        )

        tmin_original = swath_data.time[0]
        tmax_original = swath_data.time[-1]

        if from_track_range is not None:
            if isinstance(from_track_range, Iterable) and len(from_track_range) == 2:
                from_track_range = list(from_track_range)
                for i in [0, -1]:
                    if from_track_range[i] is None:
                        from_track_range[i] = swath_data.across_track_distance[i]
            swath_data = swath_data.select_from_track_range(from_track_range)
        else:
            from_track_range = (
                swath_data.across_track_distance[0],
                swath_data.across_track_distance[-1],
            )

        if time_range is not None:
            if isinstance(time_range, Iterable) and len(time_range) == 2:
                time_range = list(time_range)
                for i in [0, -1]:
                    if time_range[i] is None:
                        time_range[i] = to_timestamp(swath_data.time[i])
                    else:
                        time_range[i] = to_timestamp(time_range[i])
            swath_data = swath_data.select_time_range(time_range)
        else:
            time_range = (swath_data.time[0], swath_data.time[-1])

        values = swath_data.values
        time = swath_data.time
        latitude = swath_data.latitude
        longitude = swath_data.longitude
        across_track_distance = swath_data.across_track_distance
        from_track_distance = swath_data.from_track_distance
        label = swath_data.label
        units = swath_data.units
        nadir_index = swath_data.nadir_index

        self.ax_style_y = ax_style_y or self.ax_style_y
        if self.ax_style_y == "from_track_distance":
            ydata = from_track_distance
            ylabel = "Distance from track"
        elif self.ax_style_y == "across_track_distance":
            ydata = across_track_distance
            ylabel = "Distance"
        elif self.ax_style_y == "pixel":
            ydata = np.arange(len(from_track_distance))
            ylabel = "Pixel"
        ynadir = ydata[nadir_index]

        tmin = np.datetime64(time_range[0])
        tmax = np.datetime64(time_range[1])

        if len(values.shape) == 3 and values.shape[2] == 3:
            mesh = self.ax.pcolormesh(
                time,
                ydata,
                values,
                rasterized=True,
                **kwargs,
            )
        else:
            mesh = self.ax.pcolormesh(
                time,
                ydata,
                values.T,
                norm=norm,
                cmap=cmap,
                rasterized=True,
                **kwargs,
            )

            if colorbar:
                cb_kwargs = dict(
                    label=format_var_label(label, units, label_len=label_length),
                    position=colorbar_position,
                    alignment=colorbar_alignment,
                    width=colorbar_width,
                    spacing=colorbar_spacing,
                    length_ratio=colorbar_length_ratio,
                    label_outside=colorbar_label_outside,
                    ticks_outside=colorbar_ticks_outside,
                    ticks_both=colorbar_ticks_both,
                )
                if cmap.categorical:
                    self.colorbar = add_colorbar(
                        fig=self.fig,
                        ax=self.ax,
                        data=mesh,
                        cmap=cmap,
                        **cb_kwargs,  # type: ignore
                    )
                else:
                    self.colorbar = add_colorbar(
                        fig=self.fig,
                        ax=self.ax,
                        data=mesh,
                        ticks=colorbar_ticks,
                        tick_labels=colorbar_tick_labels,
                        **cb_kwargs,  # type: ignore
                    )

        if selection_time_range is not None:
            self.selection_time_range = validate_time_range(selection_time_range)

            if selection_highlight:
                if selection_highlight_inverted:
                    self.ax.axvspan(  # type: ignore
                        tmin,  # type: ignore
                        self.selection_time_range[0],  # type: ignore
                        color=selection_highlight_color,  # type: ignore
                        alpha=selection_highlight_alpha,  # type: ignore
                    )  # type: ignore
                    self.ax.axvspan(  # type: ignore
                        self.selection_time_range[1],  # type: ignore
                        tmax,  # type: ignore
                        color=selection_highlight_color,  # type: ignore
                        alpha=selection_highlight_alpha,  # type: ignore
                    )  # type: ignore
                else:
                    self.ax.axvspan(  # type: ignore
                        self.selection_time_range[0],  # type: ignore
                        self.selection_time_range[1],  # type: ignore
                        color=selection_highlight_color,  # type: ignore
                        alpha=selection_highlight_alpha,  # type: ignore
                    )  # type: ignore

            for t in self.selection_time_range:
                self.ax.axvline(  # type: ignore
                    x=t,  # type: ignore
                    color=selection_color,  # type: ignore
                    linestyle=selection_linestyle,  # type: ignore
                    linewidth=selection_linewidth,  # type: ignore
                )  # type: ignore

        if show_nadir:
            nadir_color = Color.from_optional(nadir_color)
            nadir_color_shade = "white"
            if isinstance(nadir_color, Color):
                nadir_color_shade = nadir_color.get_best_bw_contrast_color()
            nadir_line_shade = self.ax.axhline(
                y=ynadir,
                color=nadir_color_shade,
                linestyle="solid",
                linewidth=nadir_linewidth * 2,
                alpha=0.3,
                zorder=10,
            )
            nadir_line = self.ax.axhline(
                y=ynadir,
                color=Color.from_optional(nadir_color),
                linestyle="dashed",
                linewidth=nadir_linewidth,
                zorder=10,
            )

        self.ax.set_xlim((tmin, tmax))  # type: ignore
        # self.ax.set_ylim((hmin, hmax))

        self.ax_right = self.ax.twinx()
        self.ax_right.set_ylim(self.ax.get_ylim())

        self.ax_top = self.ax.twiny()
        self.ax_top.set_xlim(self.ax.get_xlim())

        if self.ax_style_y == "pixel":
            format_height_ticks(self.ax, label=ylabel, show_units=False)
        else:
            format_height_ticks(self.ax, label=ylabel)
        format_height_ticks(
            self.ax_right, show_tick_labels=False, show_units=False, label=""
        )

        if ax_style_top is not None:
            self.ax_style_top = AlongTrackAxisStyle.from_input(self.ax_style_top)
        if ax_style_bottom is not None:
            self.ax_style_bottom = AlongTrackAxisStyle.from_input(self.ax_style_bottom)

        format_along_track_axis(
            self.ax,
            self.ax_style_bottom,
            time,
            tmin,
            tmax,
            tmin_original,
            tmax_original,
            longitude[:, nadir_index],
            latitude[:, nadir_index],
            num_ticks=self.num_ticks,
        )
        format_along_track_axis(
            self.ax_top,
            self.ax_style_top,
            time,
            tmin,
            tmax,
            tmin_original,
            tmax_original,
            longitude[:, nadir_index],
            latitude[:, nadir_index],
            num_ticks=self.num_ticks,
        )

        return self

    def plot_contour(
        self,
        values: NDArray,
        time: NDArray,
        latitude: NDArray,
        longitude: NDArray,
        nadir_index: int,
        label_levels: list | NDArray | None = None,
        label_format: str | None = None,
        levels: list | NDArray | None = None,
        linewidths: int | float | list | NDArray | None = 1.5,
        linestyles: str | list | NDArray | None = "solid",
        colors: Color | str | list | NDArray | None = "black",
        zorder: int | float | None = 2,
        show_labels: bool = True,
    ) -> "SwathFigure":
        """Adds contour lines to the plot."""
        values = np.asarray(values)
        time = np.asarray(time)
        latitude = np.asarray(latitude)
        longitude = np.asarray(longitude)

        swath_data = SwathData(
            values=values,
            time=time,
            latitude=latitude,
            longitude=longitude,
            nadir_index=nadir_index,
        )

        if isinstance(colors, str):
            colors = Color.from_optional(colors)
        elif isinstance(colors, (Iterable, np.ndarray)):
            colors = [Color.from_optional(c) for c in colors]
        else:
            colors = Color.from_optional(colors)

        values = swath_data.values
        time = swath_data.time
        latitude = swath_data.latitude
        longitude = swath_data.longitude
        across_track_distance = swath_data.across_track_distance
        from_track_distance = swath_data.from_track_distance
        # label = swath_data.label
        # units = swath_data.units
        nadir_index = swath_data.nadir_index

        if self.ax_style_y == "from_track_distance":
            ydata = from_track_distance
        elif self.ax_style_y == "across_track_distance":
            ydata = across_track_distance
        elif self.ax_style_y == "pixel":
            ydata = np.arange(len(from_track_distance))

        x = time
        y = ydata
        z = values.T

        if len(y.shape) == 2:
            y = y[len(y) // 2]

        cn = self.ax.contour(
            x,
            y,
            z,
            levels=levels,
            linewidths=linewidths,
            colors=colors,
            linestyles=linestyles,
            zorder=zorder,
        )

        if show_labels:
            labels: Iterable[float]
            if label_levels:
                labels = [l for l in label_levels if l in cn.levels]
            else:
                labels = cn.levels

            cl = self.ax.clabel(
                cn,
                labels,  # type: ignore
                inline=True,
                fmt=label_format,
                fontsize="small",
                zorder=zorder,
            )

            bold_font = font_manager.FontProperties(weight="bold")
            for text in cl:
                text.set_fontproperties(bold_font)

            for l in cn.labelTexts:
                l.set_rotation(0)

        return self

    def ecplot_coastline(
        self,
        ds: xr.Dataset,
        var: str = "land_flag",
        *,
        time_var: str = TIME_VAR,
        lat_var: str = SWATH_LAT_VAR,
        lon_var: str = SWATH_LON_VAR,
        color: ColorLike = "#F3E490",
        linewidth: float | int = 0.5,
    ):
        return self.plot_contour(
            values=ds[var].values,
            time=ds[time_var].values,
            latitude=ds[lat_var].values,
            longitude=ds[lon_var].values,
            nadir_index=int(ds.nadir_index.values),
            levels=[0, 1],
            colors=Color.from_optional(color),
            show_labels=False,
            linewidths=linewidth,
        )

    def ecplot(
        self,
        ds: xr.Dataset,
        var: str,
        *,
        time_var: str = TIME_VAR,
        lat_var: str = SWATH_LAT_VAR,
        lon_var: str = SWATH_LON_VAR,
        values: NDArray | None = None,
        time: NDArray | None = None,
        nadir_index: int | None = None,
        latitude: NDArray | None = None,
        longitude: NDArray | None = None,
        show_info: bool = True,
        info_text_loc: str | None = None,
        # Common args for wrappers
        value_range: ValueRangeLike | Literal["default"] | None = "default",
        log_scale: bool | None = None,
        norm: Normalize | None = None,
        time_range: TimeRangeLike | None = None,
        from_track_range: DistanceRangeLike | None = None,
        label: str | None = None,
        units: str | None = None,
        cmap: str | Colormap | None = None,
        colorbar: bool = True,
        colorbar_ticks: ArrayLike | None = None,
        colorbar_tick_labels: ArrayLike | None = None,
        colorbar_position: str | Literal["left", "right", "top", "bottom"] = "right",
        colorbar_alignment: str | Literal["left", "center", "right"] = "center",
        colorbar_width: float = DEFAULT_COLORBAR_WIDTH,
        colorbar_spacing: float = 0.2,
        colorbar_length_ratio: float | str = "100%",
        colorbar_label_outside: bool = True,
        colorbar_ticks_outside: bool = True,
        colorbar_ticks_both: bool = False,
        selection_time_range: TimeRangeLike | None = None,
        selection_color: str | None = Color("ec:earthcare"),
        selection_linestyle: str | None = "dashed",
        selection_linewidth: float | int | None = 2.5,
        selection_highlight: bool = False,
        selection_highlight_inverted: bool = True,
        selection_highlight_color: str = Color("white"),
        selection_highlight_alpha: float = 0.5,
        ax_style_top: AlongTrackAxisStyle | str | None = None,
        ax_style_bottom: AlongTrackAxisStyle | str | None = None,
        ax_style_y: Literal[
            "from_track_distance", "across_track_distance", "pixel"
        ] = "from_track_distance",
        show_nadir: bool = True,
        nadir_color: ColorLike | None = "black",
        nadir_linewidth: int | float = 1.5,
        label_length: int = 25,
        **kwargs,
    ) -> "SwathFigure":
        # Collect all common args for wrapped plot function call
        local_args = locals()
        # Delete all args specific to this wrapper function
        del local_args["self"]
        del local_args["ds"]
        del local_args["var"]
        del local_args["time_var"]
        del local_args["lat_var"]
        del local_args["lon_var"]
        del local_args["show_info"]
        del local_args["info_text_loc"]
        # Delete kwargs to then merge it with the residual common args
        del local_args["kwargs"]
        all_args = {**local_args, **kwargs}

        if all_args["values"] is None:
            all_args["values"] = ds[var].values
        if all_args["time"] is None:
            all_args["time"] = ds[time_var].values
        if all_args["nadir_index"] is None:
            all_args["nadir_index"] = get_nadir_index(ds)
        if all_args["latitude"] is None:
            all_args["latitude"] = ds[lat_var].values
        if all_args["longitude"] is None:
            all_args["longitude"] = ds[lon_var].values

        # Set default values depending on variable name
        if label is None:
            all_args["label"] = (
                "Values" if not hasattr(ds[var], "long_name") else ds[var].long_name
            )
        if units is None:
            all_args["units"] = "-" if not hasattr(ds[var], "units") else ds[var].units
        if isinstance(value_range, str) and value_range == "default":
            value_range = None
            all_args["value_range"] = None
            if log_scale is None and norm is None:
                all_args["norm"] = get_default_norm(var, file_type=ds)
        if cmap is None:
            all_args["cmap"] = get_default_cmap(var, file_type=ds)

        ds = ensure_updated_msi_rgb_if_required(ds, var, time_range, time_var=time_var)

        self.plot(**all_args)

        self._set_info_text_loc(info_text_loc)
        if show_info:
            self.info_text = add_text_product_info(
                self.ax, ds, append_to=self.info_text, loc=self.info_text_loc
            )

        return self

    def to_texture(self) -> "SwathFigure":
        """Convert the figure to a texture by removing all axis ticks, labels, annotations, and text."""
        # Remove anchored text and other artist text objects
        for artist in reversed(self.ax.artists):
            if isinstance(artist, (Text, AnchoredOffsetbox)):
                artist.remove()

        # Completely remove axis ticks and labels
        self.ax.axis("off")

        if self.ax_top:
            self.ax_top.axis("off")

        if self.ax_right:
            self.ax_right.axis("off")

        # Remove white frame around figure
        self.fig.subplots_adjust(left=0, right=1, bottom=0, top=1)

        # Remove colorbar
        if self.colorbar:
            self.colorbar.remove()

        return self

    def invert_xaxis(self) -> "SwathFigure":
        """Invert the x-axis."""
        self.ax.invert_xaxis()
        if self.ax_top:
            self.ax_top.invert_xaxis()
        return self

    def invert_yaxis(self) -> "SwathFigure":
        """Invert the y-axis."""
        self.ax.invert_yaxis()
        if self.ax_right:
            self.ax_right.invert_yaxis()
        return self

    def set_colorbar_tick_scale(
        self,
        multiplier: float | None = None,
        fontsize: float | str | None = None,
    ) -> "SwathFigure":
        _cb = self.colorbar
        cb: Colorbar
        if isinstance(_cb, Colorbar):
            cb = _cb
        else:
            return self

        if fontsize is not None:
            cb.ax.tick_params(labelsize=fontsize)
            return self

        if multiplier is not None:
            _fontsize = cb.ax.yaxis.get_ticklabels()[0].get_fontsize()
            if isinstance(_fontsize, str):
                fp = font_manager.FontProperties(size=_fontsize)
                _fontsize = fp.get_size_in_points()
            cb.ax.tick_params(labelsize=_fontsize * multiplier)
        return self

    def show(self) -> None:
        import IPython
        import matplotlib.pyplot as plt
        from IPython.display import display

        if IPython.get_ipython() is not None:
            display(self.fig)
        else:
            plt.show()

    def save(
        self,
        filename: str = "",
        filepath: str | None = None,
        ds: xr.Dataset | None = None,
        ds_filepath: str | None = None,
        dpi: float | Literal["figure"] = "figure",
        orbit_and_frame: str | None = None,
        utc_timestamp: TimestampLike | None = None,
        use_utc_creation_timestamp: bool = False,
        site_name: str | None = None,
        hmax: int | float | None = None,
        radius: int | float | None = None,
        extra: str | None = None,
        transparent_outside: bool = False,
        verbose: bool = True,
        print_prefix: str = "",
        create_dirs: bool = False,
        transparent_background: bool = False,
        resolution: str | None = None,
        **kwargs,
    ) -> None:
        """
        Save a figure as an image or vector graphic to a file and optionally format the file name in a structured way using EarthCARE metadata.

        Args:
            figure (Figure | HasFigure): A figure object (`matplotlib.figure.Figure`) or objects exposing a `.fig` attribute containing a figure (e.g., `CurtainFigure`).
            filename (str, optional): The base name of the file. Can be extended based on other metadata provided. Defaults to empty string.
            filepath (str | None, optional): The path where the image is saved. Can be extended based on other metadata provided. Defaults to None.
            ds (xr.Dataset | None, optional): A EarthCARE dataset from which metadata will be taken. Defaults to None.
            ds_filepath (str | None, optional): A path to a EarthCARE product from which metadata will be taken. Defaults to None.
            pad (float, optional): Extra padding (i.e., empty space) around the image in inches. Defaults to 0.1.
            dpi (float | 'figure', optional): The resolution in dots per inch. If 'figure', use the figure's dpi value. Defaults to None.
            orbit_and_frame (str | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            utc_timestamp (TimestampLike | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            use_utc_creation_timestamp (bool, optional): Whether the time of image creation should be included in the file name. Defaults to False.
            site_name (str | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            hmax (int | float | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            radius (int | float | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            resolution (str | None, optional): Metadata used in the formatting of the file name. Defaults to None.
            extra (str | None, optional): A custom string to be included in the file name. Defaults to None.
            transparent_outside (bool, optional): Whether the area outside figures should be transparent. Defaults to False.
            verbose (bool, optional): Whether the progress of image creation should be printed to the console. Defaults to True.
            print_prefix (str, optional): A prefix string to all console messages. Defaults to "".
            create_dirs (bool, optional): Whether images should be saved in a folder structure based on provided metadata. Defaults to False.
            transparent_background (bool, optional): Whether the background inside and outside of figures should be transparent. Defaults to False.
            **kwargs (dict[str, Any]): Keyword arguments passed to wrapped function call of `matplotlib.pyplot.savefig`.
        """
        save_plot(
            fig=self.fig,
            filename=filename,
            filepath=filepath,
            ds=ds,
            ds_filepath=ds_filepath,
            dpi=dpi,
            orbit_and_frame=orbit_and_frame,
            utc_timestamp=utc_timestamp,
            use_utc_creation_timestamp=use_utc_creation_timestamp,
            site_name=site_name,
            hmax=hmax,
            radius=radius,
            extra=extra,
            transparent_outside=transparent_outside,
            verbose=verbose,
            print_prefix=print_prefix,
            create_dirs=create_dirs,
            transparent_background=transparent_background,
            resolution=resolution,
            **kwargs,
        )



import numpy as np
from numpy.typing import ArrayLike, NDArray


def normalize_longitude(lons: ArrayLike) -> NDArray:
    """Ensures -180/+180 longitude range."""
    return ((np.array(lons) + 180) % 360) - 180



from typing import Any, Literal, Sequence, Type, TypeAlias

import numpy as np
import pandas as pd
from matplotlib.axes import Axes
from numpy.typing import NDArray

from ..color.color import Color, ColorLike

ClassIDInt: TypeAlias = int
ClassLabelStr: TypeAlias = str
import matplotlib.pyplot as plt


def plot_stacked_propabilities(
    ax: Axes,
    probabilities: NDArray,
    time: NDArray | None = None,
    colors: Sequence | NDArray | None = None,
    labels: Sequence | NDArray | None = None,
    zorder: int | float = 2,
    ax_label: str | None = None,
) -> None:
    x: NDArray = np.array(range(probabilities.shape[0]))
    if isinstance(time, (Sequence, np.ndarray)):
        x = np.array(time)

    ys = tuple(probabilities[:, i] for i in range(probabilities.shape[1]))
    sp = ax.stackplot(
        x,
        ys,
        colors=colors,
        linewidth=0,
        zorder=zorder,
    )
    ax.margins(0)

    colors = [Color(obj.get_facecolor()[0]).hex for obj in sp]  # type: ignore

    if labels is not None:
        ax.legend(
            labels=labels[::-1],
            labelcolor=colors[::-1],
            handles=[ax.plot([], [])[0] for _ in labels],
            handlelength=0,
            handleheight=0,
            handletextpad=0,
            bbox_to_anchor=(1, 0.5),
            fancybox=False,
            loc="center left",
            borderaxespad=0,
            frameon=False,
        )

    if isinstance(ax_label, str):
        ax.set_ylabel(ax_label)



import numpy as np


def validate_orbit_number(orbit_number: int | str) -> int:
    """Checks if given orbit number is valid and raises ValueError if the given number is invalid or TypeError if type is wrong."""
    if isinstance(orbit_number, (int, np.integer)) or isinstance(orbit_number, str):
        value_error_msg = f"Given orbit number is invalid: {orbit_number}"
        try:
            orbit_number = int(orbit_number)
        except ValueError:
            raise ValueError(value_error_msg)

        if orbit_number < 0 or orbit_number > 99999:
            raise ValueError(value_error_msg)

        return orbit_number
    raise TypeError(
        f"Given orbit number has invalid type ({type(orbit_number)}: {orbit_number})"
    )


def format_orbit_number(orbit_number: int | str | None) -> str:
    """Returns validated orbit number as a 5 digit string or regex string '.....' if None."""
    if orbit_number is None:
        return "....."
    orbit_number = validate_orbit_number(orbit_number)
    return str(orbit_number).zfill(5)


def validate_frame_id(frame_id: str) -> str:
    """Checks if given frame ID is valid and raises ValueError if the given charater is invalid or TypeError if type is wrong."""
    if isinstance(frame_id, str):
        if frame_id.isalpha() and frame_id.upper() in [
            "A",
            "B",
            "C",
            "D",
            "E",
            "F",
            "G",
            "H",
        ]:
            return frame_id.upper()
        raise ValueError(f"Given frame ID is invalid: '{frame_id}'")
    else:
        raise TypeError(
            f"Given frame ID has invalid type ({type(frame_id)}: {frame_id})"
        )


def format_frame_id(frame_id: str | None) -> str:
    """Returns validated frame ID as one letter string or regex string '.' if None."""
    if frame_id is None:
        return "."
    return validate_frame_id(frame_id)


def format_orbit_and_frame(orbit_and_frame: str) -> str:
    """Returns validated and formatted orbit and frame 6 character string."""
    if isinstance(orbit_and_frame, str):
        try:
            if len(orbit_and_frame) == 1 and orbit_and_frame.isalpha():
                orbit_number = None
                frame_id = orbit_and_frame
            else:
                orbit_and_frame = orbit_and_frame.upper().zfill(6)
                frame_id = orbit_and_frame[-1]
                if str(frame_id).isalpha():
                    orbit_number = orbit_and_frame[:-1]
                else:
                    frame_id = None
                    orbit_number = orbit_and_frame
            frame_id = format_frame_id(frame_id)
            orbit_number = format_orbit_number(orbit_number)
            orbit_and_frame = f"{orbit_number}{frame_id}"
            return orbit_and_frame
        except ValueError:
            raise ValueError(f"Given orbit and frame is invalid: {orbit_and_frame}")
    else:
        raise TypeError(
            f"Given orbit and frame has invalid type ({type(orbit_and_frame)}: {orbit_and_frame})"
        )



from .concat import concat_datasets
from .delete import remove_dims
from .exception import EmptyFilterResultError
from .filter_index import filter_index
from .filter_latitude import filter_latitude
from .filter_radius import filter_radius
from .filter_time import filter_time, get_filter_time_mask
from .insert_var import insert_var
from .merge import merge_datasets
from .scalars import convert_scalar_var_to_str

__all__ = [
    "concat_datasets",
    "remove_dims",
    "EmptyFilterResultError",
    "filter_index",
    "filter_latitude",
    "filter_radius",
    "filter_time",
    "get_filter_time_mask",
    "insert_var",
    "merge_datasets",
    "convert_scalar_var_to_str",
]



from typing import Final, TypeAlias

from ._eo_collection import EOCollection
from ._eo_parameters import get_available_parameters

CollectionStr: TypeAlias = str
ProductTypeStr: TypeAlias = str

COLLECTION_PRODUCT_TYPE_DICT: Final[dict[CollectionStr, list[ProductTypeStr]]] = {
    "EarthCAREOrbitData": [],
    "EarthCAREXMETL1DProducts10": ["AUX_MET_1D"],
    "EarthCAREL2Products": [
        "ACM_CAP_2B",
        "ACM_COM_2B",
        "ACM_RT__2B",
        "AC__TC__2B",
        "ALL_3D__2B",
        "ALL_DF__2B",
        "AM__ACD_2B",
        "AM__CTH_2B",
        "ATL_AER_2A",
        "ATL_ALD_2A",
        "ATL_CTH_2A",
        "ATL_EBD_2A",
        "ATL_FM__2A",
        "ATL_ICE_2A",
        "ATL_TC__2A",
        "BMA_FLX_2B",
        "BM__RAD_2B",
        "CPR_APC_2A",
        "CPR_CD__2A",
        "CPR_CLD_2A",
        "CPR_FMR_2A",
        "CPR_TC__2A",
        "MSI_AOT_2A",
        "MSI_CM__2A",
        "MSI_COP_2A",
    ],
    "EarthCAREL2Validated": [
        "AC__TC__2B",
        "AM__ACD_2B",
        "AM__CTH_2B",
        "ATL_AER_2A",
        "ATL_ALD_2A",
        "ATL_CTH_2A",
        "ATL_EBD_2A",
        "ATL_FM__2A",
        "ATL_ICE_2A",
        "ATL_TC__2A",
        "BM__RAD_2B",
        "CPR_CD__2A",
        "CPR_CLD_2A",
        "CPR_FMR_2A",
        "CPR_TC__2A",
        "MSI_AOT_2A",
        "MSI_CM__2A",
        "MSI_COP_2A",
    ],
    "EarthCAREL2Validated_MAAP": [
        "AC__TC__2B",
        "AM__ACD_2B",
        "AM__CTH_2B",
        "ATL_AER_2A",
        "ATL_ALD_2A",
        "ATL_CTH_2A",
        "ATL_EBD_2A",
        "ATL_FM__2A",
        "ATL_ICE_2A",
        "ATL_TC__2A",
        "BM__RAD_2B",
        "CPR_CD__2A",
        "CPR_CLD_2A",
        "CPR_FMR_2A",
        "CPR_TC__2A",
        "MSI_AOT_2A",
        "MSI_CM__2A",
        "MSI_COP_2A",
    ],
    "EarthCAREL2Products_MAAP": [
        "AC__TC__2B",
        "AM__ACD_2B",
        "AM__CTH_2B",
        "ATL_AER_2A",
        "ATL_ALD_2A",
        "ATL_CTH_2A",
        "ATL_EBD_2A",
        "ATL_FM__2A",
        "ATL_ICE_2A",
        "ATL_TC__2A",
        "BM__RAD_2B",
        "CPR_CD__2A",
        "CPR_CLD_2A",
        "CPR_FMR_2A",
        "CPR_TC__2A",
        "MSI_AOT_2A",
        "MSI_CM__2A",
        "MSI_COP_2A",
    ],
    "JAXAL2InstChecked_MAAP": [],
    "JAXAL2Products_MAAP": [],
    "EarthCAREL0L1Products": [
        "ATL_CSC_1B",
        "ATL_DCC_1B",
        "ATL_FSC_1B",
        "ATL_NOM_0_",
        "ATL_NOM_1B",
        "AUX_JSG_1D",
        "AUX_MET_1D",
        "BBR_LIN_1B",
        "BBR_NOM_0_",
        "BBR_NOM_1B",
        "BBR_SNG_1B",
        "BBR_SOL_1B",
        "CPR_NOM_1B",
        "MSI_BBS_1B",
        "MSI_DRK_1B",
        "MSI_NOM_0_",
        "MSI_NOM_1B",
        "MSI_RGR_1C",
        "MSI_SD1_1B",
        "MSI_SD2_1B",
        "MSI_TRF_1B",
        "TLM_NOM_0_",
    ],
    "EarthCAREL2InstChecked_MAAP": [
        "AC__TC__2B",
        "AM__ACD_2B",
        "AM__CTH_2B",
        "ATL_AER_2A",
        "ATL_ALD_2A",
        "ATL_CTH_2A",
        "ATL_EBD_2A",
        "ATL_FM__2A",
        "ATL_ICE_2A",
        "ATL_TC__2A",
        "BM__RAD_2B",
        "CPR_CD__2A",
        "CPR_CLD_2A",
        "CPR_FMR_2A",
        "CPR_TC__2A",
        "MSI_AOT_2A",
        "MSI_CM__2A",
        "MSI_COP_2A",
    ],
    "EarthCAREL0L1Products_MAAP": [
        "ATL_NOM_1B",
        "BBR_LIN_1B",
        "BBR_NOM_1B",
        "BBR_SNG_1B",
        "BBR_SOL_1B",
        "CPR_NOM_1B",
        "MSI_BBS_1B",
        "MSI_DRK_1B",
        "MSI_NOM_1B",
        "MSI_RGR_1C",
        "MSI_SD1_1B",
        "MSI_SD2_1B",
        "MSI_TRF_1B",
    ],
    "EarthCAREL1Validated_MAAP": [
        "ATL_NOM_1B",
        "BBR_NOM_1B",
        "BBR_SNG_1B",
        "CPR_NOM_1B",
        "MSI_NOM_1B",
        "MSI_RGR_1C",
    ],
    "EarthCAREXMETL1DProducts10_MAAP": [],
    "EarthCAREL1InstChecked_MAAP": [
        "ATL_NOM_1B",
        "BBR_NOM_1B",
        "BBR_SNG_1B",
        "CPR_NOM_1B",
        "MSI_NOM_1B",
        "MSI_RGR_1C",
    ],
    "EarthCAREL2InstChecked": [
        "ACM_CAP_2B",
        "ACM_COM_2B",
        "ACM_RT__2B",
        "AC__TC__2B",
        "ALL_3D__2B",
        "ALL_DF__2B",
        "AM__ACD_2B",
        "AM__CTH_2B",
        "ATL_AER_2A",
        "ATL_ALD_2A",
        "ATL_CTH_2A",
        "ATL_EBD_2A",
        "ATL_FM__2A",
        "ATL_ICE_2A",
        "ATL_TC__2A",
        "BMA_FLX_2B",
        "BM__RAD_2B",
        "CPR_APC_2A",
        "CPR_CD__2A",
        "CPR_CLD_2A",
        "CPR_FMR_2A",
        "CPR_TC__2A",
        "MSI_AOT_2A",
        "MSI_CM__2A",
        "MSI_COP_2A",
    ],
    "EarthCAREL1InstChecked": [
        "ATL_CSC_1B",
        "ATL_DCC_1B",
        "ATL_FSC_1B",
        "ATL_NOM_1B",
        "AUX_JSG_1D",
        "AUX_MET_1D",
        "BBR_NOM_1B",
        "BBR_SNG_1B",
        "CPR_NOM_1B",
        "MSI_NOM_1B",
        "MSI_RGR_1C",
    ],
    "JAXAL2Validated_MAAP": [],
    "EarthCAREOrbitData_MAAP": [],
    "EarthCAREAuxiliary_MAAP": [],
    "EarthCAREL1Validated": [
        "ATL_NOM_1B",
        "AUX_JSG_1D",
        "BBR_NOM_1B",
        "BBR_SNG_1B",
        "CPR_NOM_1B",
        "MSI_NOM_1B",
        "MSI_RGR_1C",
    ],
    "JAXAL2Products": [
        "ACM_CLP_2B",
        "AC__CLP_2B",
        "ALL_RAD_2B",
        "ATL_CLA_2A",
        "CPR_CLP_2A",
        "CPR_ECO_2A",
        "MSI_CLP_2A",
    ],
    "JAXAL2Validated": [],
    "EarthCAREAuxiliary": [
        "AUX_CCAM_A",
        "AUX_CCAR_A",
        "AUX_CCAX_A",
        "AUX_CCDC_A",
        "AUX_CCER_B",
        "AUX_CCGT_B",
        "AUX_CCIA_A",
        "AUX_CCIA_B",
        "AUX_CCIA_M",
        "AUX_CCLI_B",
        "AUX_CCMC_A",
        "AUX_CCOG_A",
        "AUX_CCOG_B",
        "AUX_CCPS_A",
        "AUX_CCPS_B",
        "AUX_CCPS_M",
        "AUX_CCSN_B",
        "AUX_CCSX_A",
        "AUX_CSMO_M",
        "AUX_ORBPRE",
        "AUX_ORBRES",
        "AUX_STAT_M",
        "AUX_TBBL_M",
        "AUX_TSCO_M",
        "AUX_VRAD_M",
        "GEO_ATTOBS",
        "GEO_ORBOBS",
        "MPL_ORBSCT",
    ],
    "JAXAL2InstChecked": [],
}
"""To reduce number of performed API request, this variable is by default hardcoded (last updated on 2025-07-16)."""


def get_collection_product_type_dict(
    collections: list[EOCollection], perform_requests: bool = False
) -> dict[CollectionStr, list[ProductTypeStr]]:
    """Retruns a dictionary mapping collection names to lists of available products."""

    if not perform_requests:
        return COLLECTION_PRODUCT_TYPE_DICT

    collection_product_type_dict: dict[CollectionStr, list[ProductTypeStr]] = {}
    for c in collections:
        params = get_available_parameters(c)
        if len(params) == 0:
            collection_product_type_dict[c.name] = []
            continue

        product_type = [p for p in params if p.name == "productType"]
        if len(product_type) == 0:
            collection_product_type_dict[c.name] = []
            continue

        product_type_enum = product_type[0].enum
        if not isinstance(product_type_enum, list):
            collection_product_type_dict[c.name] = []
            continue

        collection_product_type_dict[c.name] = product_type_enum
    return collection_product_type_dict



from ._compare_bsc_ext_lr_depol import compare_bsc_ext_lr_depol
from ._perform_anom_depol_statistics import perform_anom_depol_statistics

__all__ = [
    "compare_bsc_ext_lr_depol",
    "perform_anom_depol_statistics",
]



from typing import Final

# Dataset dimension names
ALONG_TRACK_DIM: Final[str] = "along_track"
ACROSS_TRACK_DIM: Final[str] = "across_track"
VERTICAL_DIM: Final[str] = "vertical"

# Dataset variable names
TIME_VAR: Final[str] = "time"
HEIGHT_VAR: Final[str] = "height"

TRACK_LAT_VAR: Final[str] = "latitude"
TRACK_LON_VAR: Final[str] = "longitude"

SWATH_LAT_VAR: Final[str] = "latitude_swath"
SWATH_LON_VAR: Final[str] = "longitude_swath"

ELEVATION_VAR: Final[str] = "elevation"
TROPOPAUSE_VAR: Final[str] = "tropopause_height"
TEMP_CELSIUS_VAR: Final[str] = "temperature_celsius"
TEMP_KELVIN_VAR: Final[str] = "temperature_kelvin"
PRESSURE_VAR: Final[str] = "pressure"
LAND_FLAG_VAR: Final[str] = "land_flag"

ACROSS_TRACK_DISTANCE: Final[str] = "across_track_distance"
FROM_TRACK_DISTANCE: Final[str] = "from_track_distance"

NADIR_INDEX_VAR: Final[str] = "nadir_index"

# Dataset variable labels (i.e. long_name attributes)
BSC_LABEL: Final[str] = "Bsc. coeff."
EXT_LABEL: Final[str] = "Ext. coeff."
LR_LABEL: Final[str] = "Lidar ratio"
DEPOL_LABEL: Final[str] = "Depol. ratio"

# Units
UNITS_MSI_RADIANCE: Final[str] = "Wm$^{-2}$ sr$^{-1}$ µm$^{-1}$"
UNITS_KELVIN: Final[str] = "K"



import xarray as xr


def remove_dims(ds: xr.Dataset, dims_to_remove: list[str]) -> xr.Dataset:
    """Drop a list of dimensions and all associated variables and coordinates from a given `xarray.dataset`."""
    vars_to_drop = [
        var
        for var in ds.variables
        if any(dim in ds[var].dims for dim in dims_to_remove)
    ]
    coords_to_drop = [
        coord
        for coord in ds.coords
        if any(dim in ds[coord].dims for dim in dims_to_remove)
    ]

    ds_dropped = ds.drop_vars(vars_to_drop + coords_to_drop, errors="ignore")

    for dim in dims_to_remove:
        if dim in ds_dropped.dims:
            ds_dropped = ds_dropped.drop_dims(dim)

    return ds_dropped



from ._parse_dirpaths import parse_path_to_data, parse_path_to_imgs
from ._parse_path_to_config import parse_path_to_config
from ._parse_search_inputs import parse_search_inputs
from ._parse_selection_index import parse_selected_index



from collections import defaultdict
from logging import Logger
from typing import Any

import pandas as pd

from ..utils._cli import console_exclusive_info, log_textbox
from ._eo_product import EOProduct
from ._eo_search_request import EOSearchRequest
from ._exceptions import InvalidInputError


def run_search_requets(
    search_requests: list[EOSearchRequest],
    is_debug: bool,
    is_found_files_list_to_txt: bool,
    log_heading_msg: str = f"Search products",
    selected_index_input: int | None = None,
    selected_index: int | None = None,
    logger: Logger | None = None,
    download_only_h5: bool = False,
) -> list[EOProduct]:
    if (
        isinstance(selected_index_input, int) and not isinstance(selected_index, int)
    ) or (
        not isinstance(selected_index_input, int) and isinstance(selected_index, int)
    ):
        raise KeyError(f"Missing selected_index_input or selected_index")

    if logger:
        console_exclusive_info()
        log_textbox(log_heading_msg, logger=logger, show_time=True)
        console_exclusive_info()

    total_count: int = len(search_requests)
    found_products: list[EOProduct] = []
    for i, sr in enumerate(search_requests):
        _counter: int = i + 1
        _products: list[EOProduct] = sr.run(
            counter=_counter,
            total_count=total_count,
            logger=logger,
            download_only_h5=download_only_h5,
        )
        found_products.extend(_products)

    # Drop duplicates
    found_products.sort()
    found_products = list({p.name: p for p in found_products}.values())
    # Ensure sorted list
    found_products.sort()

    total_results: int = len(found_products)

    if total_results == 0:
        if logger:
            logger.info(f"No files where found for your request")
        return []

    if logger:
        console_exclusive_info()
        logger.info(f"List of files found (total number {total_results}):")

    if isinstance(selected_index_input, int) and isinstance(selected_index, int):
        try:
            selected_product: EOProduct = found_products[selected_index]
        except IndexError:
            raise InvalidInputError(
                f"The index you selected exceeds the bounds of the found files list (1 - {total_results})"
            )

    if logger:
        max_idx_str_len = len(str(total_results))
        for i, file in enumerate(found_products):
            idx_str = str(i + 1)
            msg = f" [{idx_str.rjust(max_idx_str_len)}]  {file.name}"
            if isinstance(selected_index, int) and i == selected_index:
                msg = f"<[{idx_str.rjust(max_idx_str_len)}]> {file.name} <-- Select file (user input: {selected_index_input})"
            if total_results > 41:
                if i == 20:
                    console_exclusive_info(f" ... {total_results - 40} more files ...")
                if i < 20 or total_results - i <= 20:
                    if not is_debug:
                        console_exclusive_info(msg)
            else:
                if not is_debug:
                    console_exclusive_info(msg)
            logger.debug(msg)

        if is_found_files_list_to_txt:
            df = pd.DataFrame({"id": [p.name for p in found_products]})
            df["id"].to_csv("results.txt", index=False, header=False)
        else:
            logger.info(f"Note: To export this list use the option --export_results")

    if isinstance(selected_index, int):
        return [found_products[selected_index]]
    else:
        if logger:
            logger.info(
                f"Note: To select only one specific file use the option -i/--select_file_at_index"
            )
    return found_products



import warnings

import xarray as xr
from scipy.interpolate import interp1d  # type: ignore

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import (
    rename_and_create_temperature_vars,
    rename_common_dims_and_vars,
)
from ..file_info import FileAgency
from ..science_group import read_science_data


def read_product_xmet(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
    ensure_nans: bool = DEFAULT_READ_EC_PRODUCT_ENSURE_NANS,
    **kwargs,
) -> xr.Dataset:
    """Opens AUX_MET_1D file as a `xarray.Dataset`."""
    ds = read_science_data(
        filepath,
        agency=FileAgency.ESA,
        ensure_nans=ensure_nans,
        **kwargs,
    )

    if not modify:
        return ds

    ds = rename_common_dims_and_vars(
        ds=ds,
        tropopause_var="tropopause_height_calipso",
        temperature_var="temperature",
    )

    return ds
